import datetime
from itertools import count, tee
from typing import Sequence, Set, Union
from unittest import TestCase
from unittest.mock import Mock, MagicMock

import dateutil
from parameterized import parameterized

from benji.database import VersionUid, Version
from benji.retentionfilter import RetentionFilter


# From https://docs.python.org/3/library/itertools.html#itertools-recipes
def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


class RetentionFilterTestCase(TestCase):

    REF_TIME = datetime.datetime(2019, 5, 9, 0, 0, 0, 0, tzinfo=None)

    versions: Set[Version]

    @staticmethod
    def _make_version(uid: str, date: datetime.datetime) -> Version:
        version = MagicMock(spec=('uid', 'date', '__repr__'))
        version.uid = VersionUid(uid)
        version.date = date
        version.__repr__ = Mock()
        version.__repr__.return_value = '{} - {}'.format(version.uid, version.date.isoformat(timespec='seconds'))
        return version

    @classmethod
    def setUpClass(cls) -> None:
        timestamps = (cls.REF_TIME - dateutil.relativedelta.relativedelta(minutes=n * 15) for n in range(0, 8640 + 1))
        cls.versions = set([cls._make_version(f'v{c}', t) for c, t in zip(count(start=1), timestamps)])

    @parameterized.expand([
        ('latest3', 3, 15 * 60),
        ('latest10', 10, 15 * 60),
        ('hours12', 13, 60 * 60),
        ('hours25', 26, 60 * 60),
        ('days15', 16, 24 * 60 * 60),
        ('days31', 32, 24 * 60 * 60),
        ('weeks4', 5, 7 * 24 * 60 * 60),
        ('months1', 2, 31 * 24 * 60 * 60),
        ('months2', 3, 31 * 24 * 60 * 60),
        ('years1', 1, 366 * 31 * 24 * 60 * 60),
    ])
    def test_single(self, spec: str, expected_length: int, base_unit: float) -> None:
        filter = RetentionFilter(rules_spec=spec, reference_time=self.REF_TIME, tz=datetime.timezone.utc)
        dismissed_versions = set(filter.filter(self.versions))
        remaining_versions = self.versions - dismissed_versions

        self.assertEqual(expected_length, len(remaining_versions))

        for version in remaining_versions:
            # No version must be older than the cutoff
            cutoff = base_unit * expected_length
            self.assertLessEqual((self.REF_TIME - version.date).total_seconds(), cutoff)

        sorted_versions = sorted(remaining_versions, key=lambda version: version.date)
        for older, younger in pairwise(sorted_versions):
            # Difference between adjacent versions must be less than or at most equal to the base unit
            self.assertLessEqual((younger.date - older.date).total_seconds(), base_unit)

        # Filter remaining versions a second time, no further versions must be filtered
        dismissed_versions_2 = filter.filter(remaining_versions)
        self.assertEqual(0, len(dismissed_versions_2))

    @parameterized.expand([
        ('latest3,months2', 6, 3 * 30 * 24 * 60 * 60),
        ('latest3,hours24,days3', 29, 4 * 24 * 60 * 60),
        ('latest3,hours48,days3', 52, 4 * 24 * 60 * 60),
        ('latest3,hours48,days3,months2', 55, 3 * 30 * 24 * 60 * 60),
        ('latest3,hours48,days30,months2', 81, 3 * 30 * 24 * 60 * 60),
    ])
    def test_multiple(self, spec: str, expected_length: int, cutoff: float) -> None:
        filter = RetentionFilter(rules_spec=spec, reference_time=self.REF_TIME, tz=datetime.timezone.utc)
        dismissed_versions = set(filter.filter(self.versions))
        remaining_versions = self.versions - set(dismissed_versions)

        self.assertEqual(expected_length, len(remaining_versions))

        for version in remaining_versions:
            # No version must be older than the cutoff
            self.assertLessEqual((self.REF_TIME - version.date).total_seconds(), cutoff)

        # Filter remaining versions a second time, no further versions must be filtered
        dismissed_versions_2 = filter.filter(remaining_versions)
        self.assertEqual(0, len(dismissed_versions_2))

    def test_moving_single(self) -> None:
        current_time = self.REF_TIME
        filter = RetentionFilter(rules_spec='hours30', reference_time=current_time, tz=datetime.timezone.utc)
        dismissed_versions = set(filter.filter(self.versions))
        remaining_versions = self.versions - set(dismissed_versions)

        self.assertEqual(31, len(remaining_versions))

        # Move in steps of 1 hour into the future
        for hour in range(1, 31):
            current_time = self.REF_TIME + dateutil.relativedelta.relativedelta(hours=hour)
            filter = RetentionFilter(rules_spec='hours30', reference_time=current_time, tz=datetime.timezone.utc)
            dismissed_versions = set(filter.filter(remaining_versions))
            # We're moving in steps of one hour, so each time one version is dismissed
            self.assertEqual(len(dismissed_versions), 1)
            # The dismissed version must be older than 31 hours
            self.assertGreaterEqual((current_time - list(dismissed_versions)[0].date).total_seconds(), 31 * 60 * 60)
            remaining_versions = remaining_versions - dismissed_versions

        self.assertEqual(1, len(remaining_versions))

    @parameterized.expand([
        ('hours30', ['hours'], [range(0, 31)], [1]),
        ('latest10', ['latest'], [range(0, 1)], [10]),
        ('latest8,hours48,days7,weeks4,months2,years1', ['latest', 'hours', 'days', 'weeks', 'months', 'years'],
         [range(0, 1), range(2, 49), range(2, 8),
          range(1, 5), range(1, 3), range(0, 1)], [8, 1, 1, 1, 1, 1]),
    ])
    def test_classification(self, spec: str, expected_categories: Sequence[str],
                            expected_timecounts: Sequence[Sequence[int]], expected_in_each: Sequence[int]) -> None:
        remaining_versions = set(self.versions)
        previous_versions_by_category_remaining = None
        # From self.REF_TIME to roughly 6 months into the future
        for hour in range(0, 4464):
            current_time = self.REF_TIME + dateutil.relativedelta.relativedelta(hours=hour)
            filter = RetentionFilter(rules_spec=spec, reference_time=current_time, tz=datetime.timezone.utc)
            dismissed_versions: Union[Sequence, Set]
            dismissed_versions, versions_by_category_remaining = filter._filter(remaining_versions)
            dismissed_versions = set(dismissed_versions)
            remaining_versions = remaining_versions - dismissed_versions

            try:
                self.assertSetEqual(set(versions_by_category_remaining.keys()), set(expected_categories))

                for i, category in enumerate(expected_categories):
                    actual_timecounts_set = set(versions_by_category_remaining[category].keys())
                    expected_timecounts_set = set(expected_timecounts[i])

                    # Due to the fact that we align on the start of day, week, etc. the first (youngest) expected
                    # timecount category can be missing.
                    if len(actual_timecounts_set) != len(expected_timecounts_set):
                        expected_timecounts_set = set(list(expected_timecounts[i])[1:])

                    self.assertSetEqual(actual_timecounts_set, expected_timecounts_set, category)

                    for timecount in versions_by_category_remaining[category].keys():
                        self.assertEqual(len(versions_by_category_remaining[category][timecount]), expected_in_each[i])
                        self.assertTrue(versions_by_category_remaining[category][timecount][0] in remaining_versions)
                        self.assertTrue(versions_by_category_remaining[category][timecount][0] not in dismissed_versions)
            except AssertionError:
                print('Time at failed assertion: {}.'.format(current_time.isoformat(timespec='seconds')))
                if previous_versions_by_category_remaining:
                    print(previous_versions_by_category_remaining)
                print(versions_by_category_remaining)
                raise

            # Add four new versions
            for minute in (15, 30, 45, 60):
                remaining_versions.add(
                    self._make_version('v{}'.format(1000000 + hour * 60 + minute),
                                       current_time + dateutil.relativedelta.relativedelta(minutes=minute)))

            previous_versions_by_category_remaining = versions_by_category_remaining



"""Fix blocks primary key

Revision ID: 368014edd88c
Revises: 151248f94062
Create Date: 2019-04-01 11:27:18.183463

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = '368014edd88c'
down_revision = '151248f94062'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('blocks', schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f('pk_blocks'))
        batch_op.create_primary_key(batch_op.f('pk_blocks'), ['version_uid', 'id'])


def downgrade():
    pass



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import os
import threading
import time
from typing import Tuple, Optional, Union, Iterator

from benji.config import ConfigDict, Config
from benji.database import DereferencedBlock, Block
from benji.exception import UsageError
from benji.io.base import IOBase
from benji.jobexecutor import JobExecutor
from benji.logging import logger


class IO(IOBase):

    def __init__(self, *, config: Config, name: str, module_configuration: ConfigDict, url: str,
                 block_size: int) -> None:
        super().__init__(config=config,
                         name=name,
                         module_configuration=module_configuration,
                         url=url,
                         block_size=block_size)

        if self.parsed_url.username or self.parsed_url.password or self.parsed_url.hostname or self.parsed_url.port \
                or self.parsed_url.params or self.parsed_url.fragment or self.parsed_url.query:
            raise UsageError('The supplied URL {} is invalid.'.format(self.url))

        self._simultaneous_reads = config.get_from_dict(module_configuration, 'simultaneousReads', types=int)
        self._simultaneous_writes = config.get_from_dict(module_configuration, 'simultaneousWrites', types=int)
        self._read_executor: Optional[JobExecutor] = None
        self._write_executor: Optional[JobExecutor] = None

    def open_r(self) -> None:
        self._read_executor = JobExecutor(name='IO-Read', workers=self._simultaneous_reads, blocking_submit=False)

    def open_w(self, size: int, force: bool = False, sparse: bool = False) -> None:
        self._write_executor = JobExecutor(name='IO-Write', workers=self._simultaneous_writes, blocking_submit=True)

        if os.path.exists(self.parsed_url.path):
            if not force:
                raise FileExistsError('{} already exists. Force the restore if you want to overwrite it.'.format(self.url))
            else:
                if size > self.size():
                    raise IOError('{} is too small. Its size is {} bytes, but we need {} bytes for the restore.'.format(
                        self.url, self.size(), size))
        else:
            with open(self.parsed_url.path, 'wb') as f:
                f.seek(size - 1)
                f.write(b'\0')

    def close(self) -> None:
        if self._read_executor:
            self._read_executor.shutdown()
        if self._write_executor:
            self._write_executor.shutdown()

    def size(self) -> int:
        with open(self.parsed_url.path, 'rb') as f:
            f.seek(0, 2)  # to the end
            size = f.tell()
        return size

    def _read(self, block: DereferencedBlock) -> Tuple[DereferencedBlock, bytes]:
        offset = block.idx * self.block_size
        t1 = time.time()
        with open(self.parsed_url.path, 'rb') as f:
            f.seek(offset)
            data = f.read(block.size)
            os.posix_fadvise(f.fileno(), offset, block.size, os.POSIX_FADV_DONTNEED)
        t2 = time.time()

        if not data:
            raise EOFError('End of file reached on {} when there should be data.'.format(self.url))

        logger.debug('{} read block {} in {:.3f}s'.format(
            threading.current_thread().name,
            block.idx,
            t2 - t1,
        ))

        return block, data

    def read(self, block: Union[DereferencedBlock, Block]) -> None:
        # We do need to dereference the block outside of the closure otherwise a reference to the block will be held
        # inside of the closure leading to database troubles.
        # See https://github.com/elemental-lf/benji/issues/61.
        block_deref = block.deref()

        def job():
            return self._read(block_deref)

        assert self._read_executor is not None
        self._read_executor.submit(job)

    def read_sync(self, block: Union[DereferencedBlock, Block]) -> bytes:
        return self._read(block.deref())[1]

    def read_get_completed(self, timeout: Optional[int] = None
                          ) -> Iterator[Union[Tuple[DereferencedBlock, bytes], BaseException]]:
        assert self._read_executor is not None
        return self._read_executor.get_completed(timeout=timeout)

    def _write(self, block: DereferencedBlock, data: bytes) -> DereferencedBlock:
        offset = block.idx * self.block_size
        t1 = time.time()
        with open(self.parsed_url.path, 'rb+') as f:
            f.seek(offset)
            written = f.write(data)
            os.posix_fadvise(f.fileno(), offset, len(data), os.POSIX_FADV_DONTNEED)
        t2 = time.time()

        logger.debug('{} wrote block {} in {:.3f}s'.format(
            threading.current_thread().name,
            block.idx,
            t2 - t1,
        ))

        assert written == len(data)
        return block

    def write(self, block: Union[DereferencedBlock, Block], data: bytes) -> None:
        # We do need to dereference the block outside of the closure otherwise a reference to the block will be held
        # inside of the closure leading to database troubles.
        # See https://github.com/elemental-lf/benji/issues/61.
        block_deref = block.deref()

        def job():
            return self._write(block_deref, data)

        assert self._write_executor is not None
        self._write_executor.submit(job)

    def write_sync(self, block: Union[DereferencedBlock, Block], data: bytes) -> None:
        self._write(block.deref(), data)

    def write_get_completed(self, timeout: Optional[int] = None) -> Iterator[Union[DereferencedBlock, BaseException]]:
        assert self._write_executor is not None
        return self._write_executor.get_completed(timeout=timeout)



import logging
import urllib.error

from prometheus_client import CollectorRegistry, Gauge, pushadd_to_gateway, generate_latest

from benji.helpers.settings import prom_push_gateway, benji_instance

logger = logging.getLogger()
command_registry = CollectorRegistry()
backup_registry = CollectorRegistry()
version_status_registry = CollectorRegistry()


def push(registry: CollectorRegistry):
    if prom_push_gateway is not None and benji_instance is not None:
        logger.info(f'Pushing Prometheus metrics to gateway {prom_push_gateway}.')
        logger.debug(generate_latest(registry).decode('utf-8'))
        try:
            pushadd_to_gateway(prom_push_gateway, job=benji_instance, registry=registry)
        except urllib.error.URLError as exception:
            logger.error(f'Pushing Prometheus metrics failed with a {type(exception).__name__} exception: {str(exception)}')
            logger.error('Ignoring.')


# yapf: disable
command_start_time = Gauge('benji_command_start_time', labelnames=['command'], documentation='Start time of Benji command (time_t)', registry=command_registry)
command_completion_time = Gauge('benji_command_completion_time', labelnames=['command'], documentation='Completion time of Benji command (time_t)', registry=command_registry)
command_runtime_seconds = Gauge('benji_command_runtime_seconds', labelnames=['command'], documentation='Runtime of Benji command (seconds)', registry=command_registry)
command_status_succeeded = Gauge('benji_command_status_succeeded', labelnames=['command'], documentation='Benji command succeeded', registry=command_registry)
command_status_failed = Gauge('benji_command_status_failed', labelnames=['command'], documentation='Benji command failed', registry=command_registry)

backup_start_time = Gauge('benji_backup_start_time', labelnames=['volume'], documentation='Start time of Benji backup command (time_t)', registry=backup_registry)
backup_completion_time = Gauge('benji_backup_completion_time', labelnames=['volume'], documentation='Completion time of Benji backup command (time_t)', registry=backup_registry)
backup_runtime_seconds = Gauge('benji_backup_runtime_seconds', labelnames=['volume'], documentation='Runtime of Benji backup command (seconds)', registry=backup_registry)
backup_status_succeeded = Gauge('benji_backup_status_succeeded', labelnames=['volume'], documentation='Benji backup command succeeded', registry=backup_registry)
backup_status_failed = Gauge('benji_backup_status_failed', labelnames=['volume'], documentation='Benji backup command failed', registry=backup_registry)

invalid_versions = Gauge('benji_invalid_versions', documentation='Number of invalid backup versions', registry=version_status_registry)
older_incomplete_versions = Gauge('benji_older_incomplete_versions', documentation='Number of older incomplete versions', registry=version_status_registry)
# yapf: enable



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import logging
import logging.config
import os
import sys
import threading
import warnings
from typing import Dict

import structlog
from structlog._frames import _find_first_app_frame_and_name

from benji.exception import UsageError
from benji.formatrenderer import FormatRenderer

logger = structlog.get_logger()


def _sl_processor_add_source_context(_, __, event_dict: Dict) -> Dict:
    frame, name = _find_first_app_frame_and_name([__name__, 'logging'])
    event_dict['file'] = frame.f_code.co_filename
    event_dict['line'] = frame.f_lineno
    event_dict['function'] = frame.f_code.co_name
    return event_dict


def _sl_processor_add_process_context(_, __, event_dict: Dict) -> Dict:
    event_dict['process'] = os.getpid()
    event_dict['thread_name'] = threading.current_thread().name
    event_dict['thread_id'] = threading.get_ident()
    return event_dict


_sl_processor_timestamper = structlog.processors.TimeStamper(utc=True)

_sl_foreign_pre_chain = [
    structlog.stdlib.add_log_level,
    _sl_processor_timestamper,
    _sl_processor_add_source_context,
    _sl_processor_add_process_context,
]

_sl_processors = [
    structlog.stdlib.add_log_level,
    structlog.stdlib.PositionalArgumentsFormatter(),
    _sl_processor_timestamper,
    _sl_processor_add_source_context,
    _sl_processor_add_process_context,
    structlog.processors.StackInfoRenderer(),
    structlog.processors.format_exc_info,
    structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
]


def init_logging(*,
                 logfile: str = None,
                 console_level: str = 'INFO',
                 console_formatter: str = 'json',
                 logfile_formatter: str = 'legacy') -> None:

    logging_config: Dict = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "console-plain": {
                "()": structlog.stdlib.ProcessorFormatter,
                "processor": FormatRenderer(colors=False, fmt='{log_color}{level_uc:>8s}: {event:s}'),
                "foreign_pre_chain": _sl_foreign_pre_chain,
            },
            "console-colored": {
                "()": structlog.stdlib.ProcessorFormatter,
                "processor": FormatRenderer(colors=True, fmt='{log_color}{level_uc:>8s}: {event:s}'),
                "foreign_pre_chain": _sl_foreign_pre_chain,
            },
            "legacy": {
                "()":
                    structlog.stdlib.ProcessorFormatter,
                "processor":
                    FormatRenderer(
                        colors=False,
                        fmt='{timestamp_local_ctime} {process:d}/{thread_name:s} {file:s}:{line:d} {level_uc:s} {event:s}'),
                "foreign_pre_chain":
                    _sl_foreign_pre_chain,
            },
            "json": {
                "()": structlog.stdlib.ProcessorFormatter,
                "processor": structlog.processors.JSONRenderer(),
                "foreign_pre_chain": _sl_foreign_pre_chain,
            },
        },
        "handlers": {
            "console": {
                "level": None,  # Filled in
                "class": "logging.StreamHandler",
                "formatter": None,  # Filled in
                "stream": "ext://sys.stderr",
            },
            "file": {
                "level": None,  # Filled in
                "class": "logging.handlers.WatchedFileHandler",
                "filename": None,  # Filled in
                "formatter": None,  # Filled in
            },
        },
        "loggers": {
            "": {
                "handlers": None,  # Filled in
                "level": "DEBUG",
                "propagate": True,
            },
        }
    }

    if console_formatter not in logging_config['formatters'].keys():
        raise UsageError('Event formatter {} is unknown.'.format(console_formatter))

    if logfile_formatter not in logging_config['formatters'].keys():
        raise UsageError('Event formatter {} is unknown.'.format(logfile_formatter))

    logging_config['handlers']['console']['formatter'] = console_formatter
    logging_config['handlers']['console']['level'] = console_level

    if logfile is not None:
        logging_config['handlers']['file']['filename'] = logfile
        logging_config['handlers']['file']['level'] = min(logging.getLevelName(console_level), logging.INFO)
        logging_config['handlers']['file']['formatter'] = logfile_formatter
    else:
        del (logging_config['handlers']['file'])

    logging_config['loggers']['']['handlers'] = logging_config['handlers'].keys()

    logging.config.dictConfig(logging_config)


# Source: https://stackoverflow.com/questions/6234405/logging-uncaught-exceptions-in-python/16993115#16993115
def _handle_exception(exc_type, exc_value, exc_traceback):
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return

    logger.error("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))


sys.excepthook = _handle_exception

structlog.configure(
    processors=_sl_processors,
    context_class=dict,
    logger_factory=structlog.stdlib.LoggerFactory(),
    wrapper_class=structlog.stdlib.BoundLogger,
    cache_logger_on_first_use=True,
)

init_logging()

# silence alembic
logging.getLogger('alembic').setLevel(logging.WARN)
# silence boto3
# See https://github.com/boto/boto3/issues/521
logging.getLogger('boto3').setLevel(logging.WARN)
logging.getLogger('botocore').setLevel(logging.WARN)
logging.getLogger('nose').setLevel(logging.WARN)
# This disables ResourceWarnings from boto3 which are normal
# See: https://github.com/boto/boto3/issues/454
warnings.filterwarnings("ignore", category=ResourceWarning, message=r'unclosed.*<(?:ssl.SSLSocket|socket\.socket).*>')
# silence b2
logging.getLogger('b2').setLevel(logging.WARN)

if os.getenv('BENJI_DEBUG_SQL') == '1':
    logging.getLogger('sqlalchemy.engine').setLevel(logging.INFO)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import base64
import datetime
import json
import os
import threading
import time
from abc import ABCMeta, abstractmethod
from typing import Union, Optional, Dict, Tuple, List, Sequence, cast, Iterator, Iterable

import semantic_version
from diskcache import FanoutCache

from benji.config import Config, ConfigDict
from benji.database import VersionUid, DereferencedBlock, BlockUid, Block
from benji.exception import ConfigurationError, BenjiException
from benji.jobexecutor import JobExecutor
from benji.logging import logger
from benji.repr import ReprMixIn
from benji.storage.dicthmac import DictHMAC
from benji.transform.base import TransformBase
from benji.transform.factory import TransformFactory
from benji.utils import TokenBucket, derive_key
from benji.versions import VERSIONS


class InvalidBlockException(BenjiException, IOError):

    def __init__(self, message: str, block: DereferencedBlock) -> None:
        super().__init__(message)

        self._block = block

    @property
    def block(self) -> DereferencedBlock:
        return self._block


class BlockNotFoundError(BenjiException, IOError):

    def __init__(self, message: str, uid: BlockUid) -> None:
        super().__init__(message)

        self._uid = uid

    @property
    def uid(self) -> BlockUid:
        return self._uid


class StorageBase(ReprMixIn, metaclass=ABCMeta):

    _CHECKSUM_KEY = 'checksum'
    _CREATED_KEY = 'created'
    _MODIFIED_KEY = 'modified'
    _HMAC_KEY = 'hmac'
    _METADATA_VERSION_KEY = 'metadata_version'
    _OBJECT_SIZE_KEY = 'object_size'
    _SIZE_KEY = 'size'
    _TRANSFORMS_KEY = 'transforms'

    _META_SUFFIX = '.meta'

    def __init__(self, *, config: Config, name: str, module_configuration: ConfigDict) -> None:
        self._name = name
        self._active_transforms: List[TransformBase] = []

        active_transforms = Config.get_from_dict(module_configuration, 'activeTransforms', None, types=list)
        if active_transforms is not None:
            for transform in active_transforms:
                self._active_transforms.append(TransformFactory.get_by_name(transform))
            logger.info('Active transforms for storage {}: {}.'.format(
                name,
                ', '.join('{} ({})'.format(transform.name, transform.module) for transform in self._active_transforms)))

        simultaneous_writes = Config.get_from_dict(module_configuration, 'simultaneousWrites', types=int)
        simultaneous_reads = Config.get_from_dict(module_configuration, 'simultaneousReads', types=int)
        simultaneous_removals = Config.get_from_dict(module_configuration, 'simultaneousRemovals', types=int)
        bandwidth_read = Config.get_from_dict(module_configuration, 'bandwidthRead', types=int)
        bandwidth_write = Config.get_from_dict(module_configuration, 'bandwidthWrite', types=int)

        self._consistency_check_writes = Config.get_from_dict(module_configuration,
                                                              'consistencyCheckWrites',
                                                              False,
                                                              types=bool)

        hmac_key_encoded = Config.get_from_dict(module_configuration, 'hmac.key', None, types=str)
        hmac_key: Optional[bytes] = None
        if hmac_key_encoded is None:
            hmac_password = Config.get_from_dict(module_configuration, 'hmac.password', None, types=str)
            if hmac_password is not None:
                hmac_kdf_salt = base64.b64decode(Config.get_from_dict(module_configuration, 'hmac.kdfSalt', types=str))
                hmac_kdf_iterations = Config.get_from_dict(module_configuration, 'hmac.kdfIterations', types=int)
                hmac_key = derive_key(salt=hmac_kdf_salt,
                                      iterations=hmac_kdf_iterations,
                                      key_length=32,
                                      password=hmac_password)
        else:
            hmac_key = base64.b64decode(hmac_key_encoded)
        self._dict_hmac: Optional[DictHMAC] = None
        if hmac_key is not None:
            logger.info('Enabling HMAC object metadata integrity protection for storage {}.'.format(name))
            self._dict_hmac = DictHMAC(hmac_key=self._HMAC_KEY, secret_key=hmac_key)

        self.read_throttling = TokenBucket()
        self.read_throttling.set_rate(bandwidth_read)  # 0 disables throttling
        self.write_throttling = TokenBucket()
        self.write_throttling.set_rate(bandwidth_write)  # 0 disables throttling

        self._read_executor = JobExecutor(name='Storage-Read', workers=simultaneous_reads, blocking_submit=False)
        self._write_executor = JobExecutor(name='Storage-Write', workers=simultaneous_writes, blocking_submit=True)
        self._remove_executor = JobExecutor(name='Storage-Remove', workers=simultaneous_removals, blocking_submit=True)

    @property
    def name(self) -> str:
        return self._name

    def _build_metadata(self,
                        *,
                        size: int,
                        object_size: int,
                        transforms_metadata: List[Dict] = None,
                        checksum: str = None) -> Tuple[Dict, bytes]:

        timestamp = datetime.datetime.utcnow().isoformat(timespec='microseconds') + 'Z'
        metadata: Dict = {
            self._CREATED_KEY: timestamp,
            self._METADATA_VERSION_KEY: str(VERSIONS.object_metadata.current),
            self._MODIFIED_KEY: timestamp,
            self._OBJECT_SIZE_KEY: object_size,
            self._SIZE_KEY: size,
        }

        if checksum:
            metadata[self._CHECKSUM_KEY] = checksum

        if transforms_metadata:
            metadata[self._TRANSFORMS_KEY] = transforms_metadata

        if self._dict_hmac:
            self._dict_hmac.add_digest(metadata)

        return metadata, json.dumps(metadata, separators=(',', ':')).encode('utf-8')

    def _decode_metadata(self, *, metadata_json: bytes, key: str, data_length: int) -> Dict:
        metadata = json.loads(metadata_json.decode('utf-8'))

        if self._dict_hmac:
            self._dict_hmac.verify_digest(metadata)

        # We currently support only one object metadata version
        if self._METADATA_VERSION_KEY not in metadata:
            raise KeyError('Required object metadata key {} is missing for object {}.'.format(
                self._METADATA_VERSION_KEY, key))
        version_obj = semantic_version.Version(metadata[self._METADATA_VERSION_KEY])
        if version_obj not in VERSIONS.object_metadata.supported:
            raise ValueError('Unsupported object metadata version: "{}".'.format(str(version_obj)))

        for required_key in [self._CREATED_KEY, self._MODIFIED_KEY, self._OBJECT_SIZE_KEY, self._SIZE_KEY]:
            if required_key not in metadata:
                raise KeyError('Required object metadata key {} is missing for object {}.'.format(required_key, key))

        if data_length != metadata[self._OBJECT_SIZE_KEY]:
            raise ValueError('Length mismatch for object {}. Expected: {}, got: {}.'.format(
                key, metadata[self._OBJECT_SIZE_KEY], data_length))

        return metadata

    def _check_write(self, *, key: str, metadata_key: str, data_expected: bytes) -> None:
        data_actual = self._read_object(key)
        metadata_actual_json = self._read_object(metadata_key)

        # Return value is ignored
        self._decode_metadata(metadata_json=metadata_actual_json, key=key, data_length=len(data_actual))

        # Comparing encapsulated data here
        if data_expected != data_actual:
            raise ValueError('Written and read data of {} differ.'.format(key))

    def _write(self, block: DereferencedBlock, data: bytes) -> DereferencedBlock:
        data, transforms_metadata = self._encapsulate(data)

        metadata, metadata_json = self._build_metadata(size=block.size,
                                                       object_size=len(data),
                                                       checksum=block.checksum,
                                                       transforms_metadata=transforms_metadata)

        key = block.uid.storage_object_to_path()
        metadata_key = key + self._META_SUFFIX

        time.sleep(self.write_throttling.consume(len(data) + len(metadata_json)))
        t1 = time.time()
        try:
            self._write_object(key, data)
            self._write_object(metadata_key, metadata_json)
        except:
            try:
                self._rm_object(key)
                self._rm_object(metadata_key)
            except FileNotFoundError:
                pass
            raise
        t2 = time.time()

        logger.debug('{} wrote data of uid {} in {:.3f}s'.format(threading.current_thread().name, block.uid, t2 - t1))

        if self._consistency_check_writes:
            try:
                self._check_write(key=key, metadata_key=metadata_key, data_expected=data)
            except (KeyError, ValueError) as exception:
                raise InvalidBlockException('Check write of block {} (UID {}) failed.'.format(block.idx, block.uid),
                                            block) from exception

        return block

    def write_block_async(self, block: Union[DereferencedBlock, Block], data: bytes) -> None:
        # We do need to dereference the block outside of the closure otherwise a reference to the block will be held
        # inside of the closure leading to database troubles.
        # See https://github.com/elemental-lf/benji/issues/61.
        block_deref = block.deref()

        def job():
            return self._write(block_deref, data)

        self._write_executor.submit(job)

    def write_block(self, block: Union[DereferencedBlock, Block], data: bytes) -> None:
        self._write(block.deref(), data)

    def write_get_completed(self, timeout: int = None) -> Iterator[Union[DereferencedBlock, BaseException]]:
        return self._write_executor.get_completed(timeout=timeout)

    def _read(self, block: DereferencedBlock, metadata_only: bool) -> Tuple[DereferencedBlock, Optional[bytes], Dict]:
        key = block.uid.storage_object_to_path()
        metadata_key = key + self._META_SUFFIX
        data: Optional[bytes] = None
        try:
            t1 = time.time()
            if not metadata_only:
                data = self._read_object(key)
                data_length = len(data)
            else:
                data_length = self._read_object_length(key)
            metadata_json = self._read_object(metadata_key)
            time.sleep(self.read_throttling.consume(len(data) if data else 0 + len(metadata_json)))
            t2 = time.time()
        except FileNotFoundError as exception:
            raise InvalidBlockException(
                'Object metadata or data of block {} (UID {}) not found.'.format(block.idx, block.uid),
                block) from exception

        try:
            metadata = self._decode_metadata(metadata_json=metadata_json, key=key, data_length=data_length)
        except (KeyError, ValueError) as exception:
            raise InvalidBlockException('Object metadata of block {} (UID {}) is invalid.'.format(block.idx, block.uid),
                                        block) from exception

        if self._CHECKSUM_KEY not in metadata:
            raise InvalidBlockException(
                'Required object metadata key {} is missing for block {} (UID {}).'.format(
                    self._CHECKSUM_KEY, block.idx, block.uid), block)

        if not metadata_only and self._TRANSFORMS_KEY in metadata:
            data = self._decapsulate(data, metadata[self._TRANSFORMS_KEY])  # type: ignore

        logger.debug('{} read data of uid {} in {:.3f}s{}'.format(threading.current_thread().name, block.uid, t2 - t1,
                                                                  ' (metadata only)' if metadata_only else ''))

        return block, data, metadata

    def read_block_async(self, block: Block, metadata_only: bool = False) -> None:
        # We do need to dereference the block outside of the closure otherwise a reference to the block will be held
        # inside of the closure leading to database troubles.
        # See https://github.com/elemental-lf/benji/issues/61.
        block_deref = block.deref()

        def job():
            return self._read(block_deref, metadata_only)

        self._read_executor.submit(job)

    def read_block(self, block: Block, metadata_only: bool = False) -> Optional[bytes]:
        return self._read(block.deref(), metadata_only)[1]

    def read_get_completed(self,
                           timeout: int = None) -> Iterator[Union[Tuple[DereferencedBlock, bytes, Dict], BaseException]]:
        return self._read_executor.get_completed(timeout=timeout)

    def check_block_metadata(self, *, block: DereferencedBlock, data_length: Optional[int], metadata: Dict) -> None:
        # Existence of keys has already been checked in _decode_metadata() and _read()
        if metadata[self._SIZE_KEY] != block.size:
            raise ValueError('Mismatch between recorded block size and data length in object metadata for block {} (UID {}). '
                             'Expected: {}, got: {}.'.format(block.idx, block.uid, block.size, metadata[self._SIZE_KEY]))

        if data_length and data_length != block.size:
            raise ValueError('Mismatch between recorded block size and actual data length for block {} (UID {}). '
                             'Expected: {}, got: {}.'.format(block.idx, block.uid, block.size, data_length))

        if block.checksum != metadata[self._CHECKSUM_KEY]:
            raise ValueError('Mismatch between recorded block checksum and checksum in object metadata for block {} (UID {}). '
                             'Expected: {}, got: {}.'.format(
                                 block.idx,
                                 block.uid,
                                 cast(str, block.checksum)[:16],  # We know that block.checksum is set
                                 metadata[self._CHECKSUM_KEY][:16]))

    def _rm_block(self, uid: BlockUid) -> BlockUid:
        key = uid.storage_object_to_path()
        metadata_key = key + self._META_SUFFIX
        try:
            self._rm_object(key)
        except FileNotFoundError as exception:
            raise BlockNotFoundError('Block UID {} not found on storage.'.format(uid), uid) from exception
        finally:
            try:
                self._rm_object(metadata_key)
            except FileNotFoundError:
                pass
        return uid

    def rm_block_async(self, uid: BlockUid) -> None:

        def job():
            return self._rm_block(uid)

        self._remove_executor.submit(job)

    def rm_block(self, uid: BlockUid) -> None:
        self._rm_block(uid)

    def rm_get_completed(self, timeout: int = None) -> Iterator[Union[BlockUid, BaseException]]:
        return self._remove_executor.get_completed(timeout=timeout)

    def wait_rms_finished(self):
        self._remove_executor.wait_for_all()

    # def rm_many_blocks(self, uids: Union[Sequence[BlockUid], AbstractSet[BlockUid]]) -> List[BlockUid]:
    #     keys = [uid.storage_object_to_path() for uid in uids]
    #     metadata_keys = [key + self._META_SUFFIX for key in keys]
    #
    #     errors = self._rm_many_objects(keys)
    #     self._rm_many_objects(metadata_keys)
    #     return [cast(BlockUid, BlockUid.storage_path_to_object(error)) for error in errors]

    def list_blocks(self) -> Iterable[BlockUid]:
        keys = self._list_objects(BlockUid.storage_prefix())
        for key in keys:
            assert isinstance(key, str)
            if key.endswith(self._META_SUFFIX):
                continue
            try:
                yield BlockUid.storage_path_to_object(key)
            except (RuntimeError, ValueError):
                # Ignore any keys which don't match our pattern to account for stray objects/files
                pass

    def list_versions(self) -> Iterable[VersionUid]:
        keys = self._list_objects(VersionUid.storage_prefix())
        for key in keys:
            assert isinstance(key, str)
            if key.endswith(self._META_SUFFIX):
                continue
            try:
                yield VersionUid.storage_path_to_object(key)
            except (RuntimeError, ValueError):
                # Ignore any keys which don't match our pattern to account for stray objects/files
                pass

    def read_version(self, version_uid: VersionUid) -> str:
        key = version_uid.storage_object_to_path()
        metadata_key = key + self._META_SUFFIX
        data = self._read_object(key)
        metadata_json = self._read_object(metadata_key)

        metadata = self._decode_metadata(metadata_json=metadata_json, key=key, data_length=len(data))

        if self._TRANSFORMS_KEY in metadata:
            data = self._decapsulate(data, metadata[self._TRANSFORMS_KEY])

        if len(data) != metadata[self._SIZE_KEY]:
            raise ValueError('Length mismatch of original data for object {}. Expected: {}, got: {}.'.format(
                key, metadata[self._SIZE_KEY], len(data)))

        return data.decode('utf-8')

    def write_version(self, version_uid: VersionUid, data: str, overwrite: Optional[bool] = False) -> None:
        key = version_uid.storage_object_to_path()
        metadata_key = key + self._META_SUFFIX

        if not overwrite:
            try:
                self._read_object(key)
            except FileNotFoundError:
                pass
            else:
                raise FileExistsError('Version {} already exists in storage.'.format(version_uid))

        data_bytes = data.encode('utf-8')
        size = len(data_bytes)

        data_bytes, transforms_metadata = self._encapsulate(data_bytes)
        metadata, metadata_json = self._build_metadata(size=size,
                                                       object_size=len(data_bytes),
                                                       transforms_metadata=transforms_metadata)

        try:
            self._write_object(key, data_bytes)
            self._write_object(metadata_key, metadata_json)
        except:
            try:
                self._rm_object(key)
                self._rm_object(metadata_key)
            except FileNotFoundError:
                pass
            raise

        if self._consistency_check_writes:
            self._check_write(key=key, metadata_key=metadata_key, data_expected=data_bytes)

    def rm_version(self, version_uid: VersionUid) -> None:
        key = version_uid.storage_object_to_path()
        metadata_key = key + self._META_SUFFIX
        try:
            self._rm_object(key)
        finally:
            try:
                self._rm_object(metadata_key)
            except FileNotFoundError:
                pass

    def storage_stats(self) -> Tuple[int, int]:
        objects_count = 0
        objects_size = 0
        for key, size in cast(Iterable[Tuple[str, int]], self._list_objects(include_size=True)):
            objects_count += 1
            objects_size += size
        return objects_count, objects_size

    def _encapsulate(self, data: bytes) -> Tuple[bytes, List]:
        if self._active_transforms is not None:
            transforms_metadata = []
            for transform in self._active_transforms:
                data_encapsulated, materials = transform.encapsulate(data=data)
                if data_encapsulated:
                    transforms_metadata.append({
                        'name': transform.name,
                        'module': transform.module,
                        'materials': materials,
                    })
                    data = data_encapsulated
            return data, transforms_metadata
        else:
            return data, []

    @staticmethod
    def _decapsulate(data: bytes, transforms_metadata: Sequence[Dict]) -> bytes:
        for element in reversed(transforms_metadata):
            name = element['name']
            module = element['module']
            transform = TransformFactory.get_by_name(name)
            if transform:
                if module != transform.module:
                    raise ConfigurationError('Mismatch between object transform module and configured module for ' +
                                             '{} ({} != {})'.format(name, module, transform.module))

                data = transform.decapsulate(data=data, materials=element['materials'])
            else:
                raise IOError('Unknown transform {} in object metadata.'.format(name))
        return data

    def wait_writes_finished(self) -> None:
        self._write_executor.wait_for_all()

    def use_read_cache(self, enable: bool) -> bool:
        return False

    def close(self) -> None:
        self._read_executor.shutdown()
        self._write_executor.shutdown()
        self._remove_executor.shutdown()

    @abstractmethod
    def _write_object(self, key: str, data: bytes):
        raise NotImplementedError

    @abstractmethod
    def _read_object(self, key: str) -> bytes:
        raise NotImplementedError

    @abstractmethod
    def _read_object_length(self, key: str) -> int:
        raise NotImplementedError

    @abstractmethod
    def _rm_object(self, key: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def _list_objects(self,
                      prefix: str = None,
                      include_size: bool = False) -> Union[Iterable[str], Iterable[Tuple[str, int]]]:
        raise NotImplementedError


class ReadCacheStorageBase(StorageBase):

    def __init__(self, *, config: Config, name: str, module_configuration: ConfigDict) -> None:
        read_cache_directory = Config.get_from_dict(module_configuration, 'readCache.directory', None, types=str)
        read_cache_maximum_size = Config.get_from_dict(module_configuration, 'readCache.maximumSize', None, types=int)
        read_cache_shards = Config.get_from_dict(module_configuration, 'readCache.shards', None, types=int)

        if read_cache_directory and read_cache_maximum_size:
            os.makedirs(read_cache_directory, exist_ok=True)
            try:
                self._read_cache = FanoutCache(
                    read_cache_directory,
                    size_limit=read_cache_maximum_size,
                    shards=read_cache_shards,
                    eviction_policy='least-frequently-used',
                    statistics=1,
                )
            except Exception:
                logger.warning('Unable to enable disk based read caching. Continuing without it.')
                self._read_cache = None
            else:
                logger.debug('Disk based read caching instantiated (cache size {}, shards {}).'.format(
                    read_cache_maximum_size, read_cache_shards))
        else:
            self._read_cache = None
        self._use_read_cache = True

        # Start reader and write threads after the disk cached is created, so that they see it.
        super().__init__(config=config, name=name, module_configuration=module_configuration)

    def _read(self, block: DereferencedBlock, metadata_only: bool) -> Tuple[DereferencedBlock, Optional[bytes], Dict]:
        key = block.uid.storage_object_to_path()
        metadata_key = key + self._META_SUFFIX
        if self._read_cache is not None and self._use_read_cache:
            metadata = self._read_cache.get(metadata_key)
            if metadata and metadata_only:
                return block, None, metadata
            elif metadata:
                data = self._read_cache.get(key)
                if data:
                    return block, data, metadata

        block, data, metadata = super()._read(block, metadata_only)

        # We always put blocks into the cache even when self._use_read_cache is False
        if self._read_cache is not None:
            self._read_cache.set(metadata_key, metadata)
            if not metadata_only:
                self._read_cache.set(key, data)

        return block, data, metadata

    def use_read_cache(self, enable: bool) -> bool:
        old_value = self._use_read_cache
        self._use_read_cache = enable
        return old_value

    def close(self) -> None:
        super().close()
        if self._read_cache is not None:
            (cache_hits, cache_misses) = self._read_cache.stats()
            logger.debug('Disk based cache statistics (since cache creation): {} hits, {} misses.'.format(
                cache_hits, cache_misses))
            self._read_cache.close()



import logging
from datetime import datetime
from tempfile import NamedTemporaryFile
from typing import Dict, Any, Optional

from blinker import signal

from benji.helpers.settings import benji_log_level
from benji.helpers.utils import subprocess_run

SIGNAL_SENDER = 'ceph'
RBD_SNAP_CREATE_TIMEOUT = 30
RBD_SNAP_NAME_PREFIX = 'b-'

logger = logging.getLogger()

signal_snapshot_create_pre = signal('snapshot_create_pre')
signal_snapshot_create_post_success = signal('snapshot_create_post_success')
signal_snapshot_create_post_error = signal('snapshot_create_post_error')
signal_backup_pre = signal('backup_pre')
signal_backup_post_success = signal('on_backup_post_success')
signal_backup_post_error = signal('on_backup_post_error')


def snapshot_create(*, volume: str, pool: str, image: str, snapshot: str, context: Any = None):
    signal_snapshot_create_pre.send(SIGNAL_SENDER,
                                    volume=volume,
                                    pool=pool,
                                    image=image,
                                    snapshot=snapshot,
                                    context=context)
    try:
        subprocess_run(['rbd', 'snap', 'create', f'{pool}/{image}@{snapshot}'], timeout=RBD_SNAP_CREATE_TIMEOUT)
    except Exception as exception:
        signal_snapshot_create_post_error.send(SIGNAL_SENDER,
                                               volume=volume,
                                               pool=pool,
                                               image=image,
                                               snapshot=snapshot,
                                               context=context,
                                               exception=exception)
    else:
        signal_snapshot_create_post_success.send(SIGNAL_SENDER,
                                                 volume=volume,
                                                 pool=pool,
                                                 image=image,
                                                 snapshot=snapshot,
                                                 context=context)


def backup_initial(*,
                   volume: str,
                   pool: str,
                   image: str,
                   version_labels: Dict[str, str],
                   version_uid: Optional[str],
                   context: Any = None) -> Dict[str, str]:
    logger.info(f'Performing initial backup of {volume}:{pool}/{image}')

    now = datetime.utcnow()
    snapshot = now.strftime(RBD_SNAP_NAME_PREFIX + '%Y-%m-%dT%H:%M:%SZ')

    snapshot_create(volume=volume, pool=pool, image=image, snapshot=snapshot, context=context)
    stdout = subprocess_run(['rbd', 'diff', '--whole-object', '--format=json', f'{pool}/{image}@{snapshot}'])

    with NamedTemporaryFile(mode='w+', encoding='utf-8') as rbd_hints:
        assert isinstance(stdout, str)
        rbd_hints.write(stdout)
        rbd_hints.flush()
        benji_args = [
            'benji', '--machine-output', '--log-level', benji_log_level, 'backup', '--snapshot', snapshot,
            '--rbd-hints', rbd_hints.name
        ]
        if version_uid is not None:
            benji_args.extend(['--uid', version_uid])
        for label_name, label_value in version_labels.items():
            benji_args.extend(['--label', f'{label_name}={label_value}'])
        benji_args.extend([f'{pool}:{pool}/{image}@{snapshot}', volume])
        result = subprocess_run(benji_args, decode_json=True)
        assert isinstance(result, dict)

    return result


def backup_differential(*,
                        volume: str,
                        pool: str,
                        image: str,
                        last_snapshot: str,
                        last_version_uid: str,
                        version_labels: Dict[str, str],
                        version_uid: Optional[str],
                        context: Any = None) -> Dict[str, str]:
    logger.info(f'Performing differential backup of {volume}:{pool}/{image} from RBD snapshot" \
        "{last_snapshot} and Benji version {last_version_uid}.')

    now = datetime.utcnow()
    snapshot = now.strftime(RBD_SNAP_NAME_PREFIX + '%Y-%m-%dT%H:%M:%SZ')

    snapshot_create(volume=volume, pool=pool, image=image, snapshot=snapshot, context=context)
    stdout = subprocess_run(
        ['rbd', 'diff', '--whole-object', '--format=json', '--from-snap', last_snapshot, f'{pool}/{image}@{snapshot}'])
    subprocess_run(['rbd', 'snap', 'rm', f'{pool}/{image}@{last_snapshot}'])

    with NamedTemporaryFile(mode='w+', encoding='utf-8') as rbd_hints:
        assert isinstance(stdout, str)
        rbd_hints.write(stdout)
        rbd_hints.flush()
        benji_args = [
            'benji', '--machine-output', '--log-level', benji_log_level, 'backup', '--snapshot', snapshot,
            '--rbd-hints', rbd_hints.name, '--base-version', last_version_uid
        ]
        if version_uid is not None:
            benji_args.extend(['--uid', version_uid])
        for label_name, label_value in version_labels.items():
            benji_args.extend(['--label', f'{label_name}={label_value}'])
        benji_args.extend([f'{pool}:{pool}/{image}@{snapshot}', volume])
        result = subprocess_run(benji_args, decode_json=True)
        assert isinstance(result, dict)

    return result


def backup(*,
           volume: str,
           pool: str,
           image: str,
           version_labels: Dict[str, str] = {},
           version_uid: str = None,
           context: Any = None):
    signal_backup_pre.send(SIGNAL_SENDER,
                           volume=volume,
                           pool=pool,
                           image=image,
                           version_labels=version_labels,
                           context=context)
    version = None
    try:
        rbd_snap_ls = subprocess_run(['rbd', 'snap', 'ls', '--format=json', f'{pool}/{image}'], decode_json=True)
        assert isinstance(rbd_snap_ls, list)
        # Snapshot are sorted by their ID, so newer snapshots come last
        benjis_snapshots = [
            snapshot['name'] for snapshot in rbd_snap_ls if snapshot['name'].startswith(RBD_SNAP_NAME_PREFIX)
        ]
        if len(benjis_snapshots) == 0:
            logger.info('No previous RBD snapshot found, performing initial backup.')
            result = backup_initial(volume=volume,
                                    pool=pool,
                                    image=image,
                                    version_uid=version_uid,
                                    version_labels=version_labels,
                                    context=context)
        else:
            # Delete all snapshots except the newest
            for snapshot in benjis_snapshots[:-1]:
                logger.info(f'Deleting older RBD snapshot {pool}/{image}@{snapshot}.')
                subprocess_run(['rbd', 'snap', 'rm', f'{pool}/{image}@{snapshot}'])

            last_snapshot = benjis_snapshots[-1]
            logger.info(f'Newest RBD snapshot is {pool}/{image}@{last_snapshot}.')

            benji_ls = subprocess_run([
                'benji', '--machine-output', '--log-level', benji_log_level, 'ls',
                f'volume == "{volume}" and snapshot == "{last_snapshot}" and status == "valid"'
            ],
                                      decode_json=True)
            assert isinstance(benji_ls, dict)
            assert 'versions' in benji_ls
            assert isinstance(benji_ls['versions'], list)
            if len(benji_ls['versions']) > 0:
                assert 'uid' in benji_ls['versions'][0]
                last_version_uid = benji_ls['versions'][0]['uid']
                assert isinstance(last_version_uid, str)
                result = backup_differential(volume=volume,
                                             pool=pool,
                                             image=image,
                                             last_snapshot=last_snapshot,
                                             last_version_uid=last_version_uid,
                                             version_uid=version_uid,
                                             version_labels=version_labels,
                                             context=context)
            else:
                logger.info(f'Existing RBD snapshot {pool}/{image}@{last_snapshot} not found in Benji, deleting it and reverting to initial backup.')
                subprocess_run(['rbd', 'snap', 'rm', f'{pool}/{image}@{last_snapshot}'])
                result = backup_initial(volume=volume,
                                        pool=pool,
                                        image=image,
                                        version_uid=version_uid,
                                        version_labels=version_labels,
                                        context=context)
        assert 'versions' in result and isinstance(result['versions'], list)
        version = result['versions'][0]
    except Exception as exception:
        signal_backup_post_error.send(SIGNAL_SENDER,
                                      volume=volume,
                                      pool=pool,
                                      image=image,
                                      version_labels=version_labels,
                                      context=context,
                                      version=version,
                                      exception=exception)
    else:
        signal_backup_post_success.send(SIGNAL_SENDER,
                                        volume=volume,
                                        pool=pool,
                                        image=image,
                                        version_labels=version_labels,
                                        context=context,
                                        version=version)



"""Remove stats table

Revision ID: 151248f94062
Revises: fe79ce75cefa
Create Date: 2019-03-28 13:08:27.087076

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '151248f94062'
down_revision = 'fe79ce75cefa'
branch_labels = None
depends_on = None

version_statistics = sa.table('version_statistics', sa.Column('uid', sa.Integer(), nullable=False),
                              sa.Column('bytes_dedup', sa.BigInteger(), nullable=True),
                              sa.Column('bytes_read', sa.BigInteger(), nullable=True),
                              sa.Column('bytes_sparse', sa.BigInteger(), nullable=True),
                              sa.Column('bytes_written', sa.BigInteger(), nullable=True),
                              sa.Column('duration', sa.BigInteger(), nullable=True))

versions = sa.table('versions', sa.Column('uid', sa.Integer(), nullable=False),
                    sa.Column('bytes_dedup', sa.BigInteger(), nullable=True),
                    sa.Column('bytes_read', sa.BigInteger(), nullable=True),
                    sa.Column('bytes_sparse', sa.BigInteger(), nullable=True),
                    sa.Column('bytes_written', sa.BigInteger(), nullable=True),
                    sa.Column('duration', sa.BigInteger(), nullable=True))


def upgrade():
    with op.batch_alter_table('versions', schema=None) as batch_op:
        batch_op.add_column(sa.Column('bytes_dedup', sa.BigInteger(), nullable=True))
        batch_op.add_column(sa.Column('bytes_read', sa.BigInteger(), nullable=True))
        batch_op.add_column(sa.Column('bytes_sparse', sa.BigInteger(), nullable=True))
        batch_op.add_column(sa.Column('bytes_written', sa.BigInteger(), nullable=True))
        batch_op.add_column(sa.Column('duration', sa.BigInteger(), nullable=True))

    op.execute(versions.update().values({
        versions.c.bytes_dedup:
        sa.sql.select(columns=[version_statistics.c.bytes_dedup]).where(versions.c.uid == version_statistics.c.uid),
        versions.c.bytes_read:
        sa.sql.select(columns=[version_statistics.c.bytes_read]).where(versions.c.uid == version_statistics.c.uid),
        versions.c.bytes_sparse:
        sa.sql.select(columns=[version_statistics.c.bytes_sparse]).where(versions.c.uid == version_statistics.c.uid),
        versions.c.bytes_written:
        sa.sql.select(columns=[version_statistics.c.bytes_written]).where(versions.c.uid == version_statistics.c.uid),
        versions.c.duration:
        sa.sql.select(columns=[version_statistics.c.duration]).where(versions.c.uid == version_statistics.c.uid),
    }))

    op.drop_table('version_statistics')


def downgrade():
    pass



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import importlib
import threading
from typing import Dict, NamedTuple, Any

from benji.config import Config, ConfigList
from benji.exception import ConfigurationError, InternalError
from benji.repr import ReprMixIn
from benji.transform.factory import TransformFactory


class _StorageFactoryModule(NamedTuple):
    module: Any
    storage_id: int
    arguments: Dict[str, Any]


class StorageFactory(ReprMixIn):

    _modules: Dict[str, _StorageFactoryModule] = {}
    _local = threading.local()

    def __init__(self) -> None:
        raise InternalError('StorageFactory constructor called.')

    @classmethod
    def _import_modules(cls, config: Config, modules: ConfigList) -> None:
        for index, module_dict in enumerate(modules):
            module = Config.get_from_dict(module_dict,
                                          'module',
                                          types=str,
                                          full_name_override=modules.full_name,
                                          index=index)
            name = Config.get_from_dict(module_dict,
                                        'name',
                                        types=str,
                                        full_name_override=modules.full_name,
                                        index=index)
            storage_id = Config.get_from_dict(module_dict,
                                              'storageId',
                                              None,
                                              types=int,
                                              full_name_override=modules.full_name,
                                              index=index)
            configuration = Config.get_from_dict(module_dict,
                                                 'configuration',
                                                 None,
                                                 types=dict,
                                                 full_name_override=modules.full_name,
                                                 index=index)

            if name in cls._modules:
                raise ConfigurationError('Duplicate name "{}" in list {}.'.format(name, modules.full_name))

            module = importlib.import_module('{}.{}'.format(__package__, module))
            try:
                configuration = config.validate(module=module.__name__, config=configuration)
            except ConfigurationError as exception:
                raise ConfigurationError('Configuration for storage {} is invalid.'.format(name)) from exception
            cls._modules[name] = _StorageFactoryModule(module=module,
                                                       storage_id=storage_id,
                                                       arguments={
                                                           'config': config,
                                                           'name': name,
                                                           'module_configuration': configuration
                                                       })

    @classmethod
    def initialize(cls, config: Config) -> None:
        TransformFactory.initialize(config)

        cls._modules = {}
        storages: ConfigList = config.get('storages', types=list)
        cls._import_modules(config, storages)

    @classmethod
    def close(cls) -> None:
        instances = cls._local.__dict__.setdefault('instances', {})

        for storage in instances.values():
            storage.close()

        cls._local.instances = {}

    @classmethod
    def get_by_name(cls, name: str) -> Any:
        instances = cls._local.__dict__.setdefault('instances', {})

        if name not in instances:
            if name not in cls._modules:
                raise ConfigurationError('Storage {} is undefined.'.format(name))

            module = cls._modules[name].module
            module_arguments = cls._modules[name].arguments
            instances[name] = module.Storage(**module_arguments)

        return instances[name]

    @classmethod
    def get_modules(cls) -> Dict[str, _StorageFactoryModule]:
        return cls._modules



import base64
from hashlib import sha256
from typing import Dict, Tuple, Optional

from Crypto.PublicKey import ECC

from benji.config import Config, ConfigDict
from benji.logging import logger
from benji.transform.aes_256_gcm import Transform as TransformAES


class Transform(TransformAES):

    def __init__(self, *, config: Config, name: str, module_configuration: ConfigDict) -> None:
        ecc_key_der: str = Config.get_from_dict(module_configuration, 'eccKey', types=str)
        ecc_curve: Optional[str] = Config.get_from_dict(module_configuration, 'eccCurve', 'NIST P-384', types=str)

        ecc_key = self._unpack_envelope_key(base64.b64decode(ecc_key_der))

        if ecc_key.curve != ecc_curve:
            raise ValueError(f'Key eccKey does not match the eccCurve setting (found: {ecc_key.curve}, expected: {ecc_curve}).')

        self._ecc_key = ecc_key
        self._ecc_curve = ecc_key.curve

        point_q_len = self._ecc_key.pointQ.size_in_bytes()
        if point_q_len < self.AES_KEY_LEN:
            raise ValueError(f'Size of point Q is smaller than the AES key length, which reduces security ({point_q_len} < {self.AES_KEY_LEN}).')

        # Note: We don't actually have a "master" aes key, because the key is derived from the ECC key
        # and set before calling the parent's encapsulate/decapsulate method.
        aes_config = module_configuration.copy()
        aes_config['masterKey'] = base64.b64encode(b'\x00' * self.AES_KEY_LEN).decode('ascii')
        super().__init__(config=config, name=name, module_configuration=aes_config)

    @staticmethod
    def _pack_envelope_key(key: ECC.EccKey) -> bytes:
        return key.export_key(format='DER', compress=True)

    @staticmethod
    def _unpack_envelope_key(key: bytes) -> ECC.EccKey:
        return ECC.import_key(key)

    @staticmethod
    def _ecc_point_to_key(point: ECC.EccPoint) -> bytes:
        sha = sha256(int.to_bytes(int(point.x), point.size_in_bytes(), 'big'))
        sha.update(int.to_bytes(int(point.y), point.size_in_bytes(), 'big'))
        return sha.digest()

    def _create_envelope_key(self) -> Tuple[bytes, bytes]:
        cipher_privkey = ECC.generate(curve=self._ecc_curve)
        shared_key = self._ecc_point_to_key(self._ecc_key.pointQ * cipher_privkey.d)
        return shared_key, self._pack_envelope_key(cipher_privkey.public_key())

    def _derive_envelope_key(self, cipher_pubkey: bytes) -> bytes:
        ecc_point = self._unpack_envelope_key(cipher_pubkey)
        return self._ecc_point_to_key(ecc_point.pointQ * self._ecc_key.d)

    def encapsulate(self, *, data: bytes) -> Tuple[Optional[bytes], Optional[Dict]]:
        if self._ecc_key.has_private():
            logger.warning('ECC key loaded from config includes private key data, which is not needed for encryption.')
        return super().encapsulate(data=data)

    def decapsulate(self, *, data: bytes, materials: Dict) -> bytes:
        if not self._ecc_key.has_private():
            raise ValueError('ECC key loaded from config does not include private key data, cannot proceed.')
        return super().decapsulate(data=data, materials=materials)



from benji.utils import keys_exist


def test_keys_exist() -> None:
    l3 = {'a': 4, 'b': 5, 'c': 6, 'd': None}
    l2 = {'a': l3, 'b': 4, 'c': 3}
    l1 = {'a': 1, 'b': 2, 'c': l2}

    assert keys_exist(l1, ['a'])
    assert not keys_exist(l1, ['d'])

    assert keys_exist(l1, ['c.a'])
    assert not keys_exist(l1, ['c.z'])
    assert not keys_exist(l1, ['a.z'])
    assert not keys_exist(l1, ['z.b'])

    assert keys_exist(l1, ['c.a.a'])
    assert not keys_exist(l1, ['c.a.z'])

    assert keys_exist(l1, ['c.a.a', 'c.a', 'c'])
    assert not keys_exist(l1, ['c.a.a', 'c.a', 'c', 'z'])

    assert keys_exist(l1, ['c.a.d'])



import json
import logging
import os
import sys
from typing import List, NamedTuple, Type, Optional, Dict

from prettytable import PrettyTable

import benji.exception
from benji import __version__
from benji.benji import Benji, BenjiStore
from benji.database import Version, VersionUid
from benji.logging import logger
from benji.nbdserver import NbdServer
from benji.utils import hints_from_rbd_diff, PrettyPrint, InputValidation, random_string
from benji.versions import VERSIONS


class _ExceptionMapping(NamedTuple):
    exception: Type[BaseException]
    exit_code: int


class Commands:
    """Proxy between CLI calls and actual backup code."""

    def __init__(self, machine_output, config):
        self.machine_output = machine_output
        self.config = config

    def backup(self, version_uid: str, volume: str, snapshot: str, source: str, rbd_hints: str, base_version_uid: str,
               block_size: int, labels: List[str], storage: str) -> None:
        if version_uid is None:
            version_uid = '{}-{}'.format(volume[:248], random_string(6))
        version_uid_obj = VersionUid(version_uid)
        base_version_uid_obj = VersionUid(base_version_uid) if base_version_uid else None

        if labels:
            label_add, label_remove = InputValidation.parse_and_validate_labels(labels)
        with Benji(self.config) as benji_obj:
            hints = None
            if rbd_hints:
                logger.debug(f'Loading RBD hints from file {rbd_hints}.')
                with open(rbd_hints, 'r') as f:
                    hints = hints_from_rbd_diff(f.read())
            backup_version = benji_obj.backup(version_uid=version_uid_obj,
                                              volume=volume,
                                              snapshot=snapshot,
                                              source=source,
                                              hints=hints,
                                              base_version_uid=base_version_uid_obj,
                                              storage_name=storage,
                                              block_size=block_size)

            if labels:
                for key, value in label_add:
                    benji_obj.add_label(backup_version.uid, key, value)
                for key in label_remove:
                    benji_obj.rm_label(backup_version.uid, key)
                if label_add:
                    logger.info('Added label(s) to version {}: {}.'.format(
                        backup_version.uid, ', '.join('{}={}'.format(name, value) for name, value in label_add)))
                if label_remove:
                    logger.info('Removed label(s) from version {}: {}.'.format(backup_version.uid,
                                                                               ', '.join(label_remove)))

            if self.machine_output:
                benji_obj.export_any({'versions': [backup_version]},
                                     sys.stdout,
                                     ignore_relationships=(((Version,), ('blocks',)),))

    def restore(self, version_uid: str, destination: str, sparse: bool, force: bool, database_less: bool,
                storage: str) -> None:
        if not database_less and storage is not None:
            raise benji.exception.UsageError('Specifying a storage location is only supported for database-less restores.')

        version_uid_obj = VersionUid(version_uid)
        with Benji(self.config, in_memory_database=database_less) as benji_obj:
            if database_less:
                benji_obj.metadata_restore([version_uid_obj], storage)
            benji_obj.restore(version_uid_obj, destination, sparse, force)

    def protect(self, version_uids: List[str]) -> None:
        version_uid_objs = [VersionUid(version_uid) for version_uid in version_uids]
        with Benji(self.config) as benji_obj:
            for version_uid in version_uid_objs:
                benji_obj.protect(version_uid, protected=True)

    def unprotect(self, version_uids: List[str]) -> None:
        version_uid_objs = [VersionUid(version_uid) for version_uid in version_uids]
        with Benji(self.config) as benji_obj:
            for version_uid in version_uid_objs:
                benji_obj.protect(version_uid, protected=False)

    def rm(self, version_uids: List[str], force: bool, keep_metadata_backup: bool, override_lock: bool) -> None:
        version_uid_objs = [VersionUid(version_uid) for version_uid in version_uids]
        disallow_rm_when_younger_than_days = self.config.get('disallowRemoveWhenYounger', types=int)
        with Benji(self.config) as benji_obj:
            for version_uid in version_uid_objs:
                benji_obj.rm(version_uid,
                             force=force,
                             disallow_rm_when_younger_than_days=disallow_rm_when_younger_than_days,
                             keep_metadata_backup=keep_metadata_backup,
                             override_lock=override_lock)

    def scrub(self, version_uid: str, block_percentage: int) -> None:
        version_uid_obj = VersionUid(version_uid)
        with Benji(self.config) as benji_obj:
            try:
                benji_obj.scrub(version_uid_obj, block_percentage=block_percentage)
            except benji.exception.ScrubbingError:
                assert benji_obj is not None
                if self.machine_output:
                    benji_obj.export_any(
                        {
                            'versions': [benji_obj.get_version_by_uid(version_uid=version_uid_obj)],
                            'errors': [benji_obj.get_version_by_uid(version_uid=version_uid_obj)]
                        },
                        sys.stdout,
                        ignore_relationships=(((Version,), ('blocks',)),))
                raise
            else:
                if self.machine_output:
                    benji_obj.export_any(
                        {
                            'versions': [benji_obj.get_version_by_uid(version_uid=version_uid_obj)],
                            'errors': []
                        },
                        sys.stdout,
                        ignore_relationships=(((Version,), ('blocks',)),))

    def deep_scrub(self, version_uid: str, source: str, block_percentage: int) -> None:
        version_uid_obj = VersionUid(version_uid)
        with Benji(self.config) as benji_obj:
            try:
                benji_obj.deep_scrub(version_uid_obj, source=source, block_percentage=block_percentage)
            except benji.exception.ScrubbingError:
                assert benji_obj is not None
                if self.machine_output:
                    benji_obj.export_any(
                        {
                            'versions': [benji_obj.get_version_by_uid(version_uid=version_uid_obj)],
                            'errors': [benji_obj.get_version_by_uid(version_uid=version_uid_obj)]
                        },
                        sys.stdout,
                        ignore_relationships=(((Version,), ('blocks',)),))
                raise
            else:
                if self.machine_output:
                    benji_obj.export_any(
                        {
                            'versions': [benji_obj.get_version_by_uid(version_uid=version_uid_obj)],
                            'errors': []
                        },
                        sys.stdout,
                        ignore_relationships=(((Version,), ('blocks',)),))

    def _batch_scrub(self, method: str, filter_expression: Optional[str], version_percentage: int,
                     block_percentage: int, group_label: Optional[str]) -> None:
        with Benji(self.config) as benji_obj:
            versions, errors = getattr(benji_obj, method)(filter_expression, version_percentage, block_percentage,
                                                          group_label)
            if errors:
                if self.machine_output:
                    benji_obj.export_any({
                        'versions': versions,
                        'errors': errors,
                    },
                                         sys.stdout,
                                         ignore_relationships=[((Version,), ('blocks',))])
                raise benji.exception.ScrubbingError('One or more version had scrubbing errors: {}.'.format(', '.join(
                    version.uid for version in errors)))
            else:
                if self.machine_output:
                    benji_obj.export_any({
                        'versions': versions,
                        'errors': []
                    },
                                         sys.stdout,
                                         ignore_relationships=(((Version,), ('blocks',)),))

    def batch_scrub(self, filter_expression: Optional[str], version_percentage: int, block_percentage: int,
                    group_label: Optional[str]) -> None:
        self._batch_scrub('batch_scrub', filter_expression, version_percentage, block_percentage, group_label)

    def batch_deep_scrub(self, filter_expression: Optional[str], version_percentage: int, block_percentage: int,
                         group_label: Optional[str]) -> None:
        self._batch_scrub('batch_deep_scrub', filter_expression, version_percentage, block_percentage, group_label)

    @staticmethod
    def _ls_versions_table_output(versions: List[Version], include_labels: bool, include_stats: bool) -> None:
        tbl = PrettyTable()

        field_names = ['date', 'uid', 'volume', 'snapshot', 'size', 'block_size', 'status', 'protected', 'storage']
        if include_stats:
            field_names.extend(['read', 'written', 'deduplicated', 'sparse', 'duration'])
        if include_labels:
            field_names.append('labels')
        tbl.field_names = field_names

        tbl.align['uid'] = 'l'
        tbl.align['volume'] = 'l'
        tbl.align['snapshot'] = 'l'
        tbl.align['storage'] = 'l'
        tbl.align['size'] = 'r'
        tbl.align['block_size'] = 'r'

        tbl.align['read'] = 'r'
        tbl.align['written'] = 'r'
        tbl.align['deduplicated'] = 'r'
        tbl.align['sparse'] = 'r'
        tbl.align['duration'] = 'r'

        tbl.align['labels'] = 'l'

        for version in versions:
            row = [
                PrettyPrint.local_time(version.date),
                version.uid,
                version.volume,
                version.snapshot,
                PrettyPrint.bytes(version.size),
                PrettyPrint.bytes(version.block_size),
                version.status,
                version.protected,
                version.storage.name,
            ]

            if include_stats:
                row.extend([
                    PrettyPrint.bytes(version.bytes_read) if version.bytes_read is not None else '',
                    PrettyPrint.bytes(version.bytes_written) if version.bytes_written is not None else '',
                    PrettyPrint.bytes(version.bytes_deduplicated) if version.bytes_deduplicated is not None else '',
                    PrettyPrint.bytes(version.bytes_sparse) if version.bytes_sparse is not None else '',
                    PrettyPrint.duration(version.duration) if version.duration is not None else '',
                ])

            if include_labels:
                row.append('\n'.join(
                    sorted(['{}={}'.format(label.name, label.value) for label in version.labels.values()])))
            tbl.add_row(row)
        print(tbl)

    def ls(self, filter_expression: Optional[str], include_labels: bool, include_stats: bool) -> None:
        with Benji(self.config) as benji_obj:
            versions = benji_obj.find_versions_with_filter(filter_expression)

            if self.machine_output:
                benji_obj.export_any(
                    {'versions': versions},
                    sys.stdout,
                    ignore_relationships=[((Version,), ('blocks',))],
                )
            else:
                self._ls_versions_table_output(versions, include_labels, include_stats)

    def cleanup(self, override_lock: bool) -> None:
        with Benji(self.config) as benji_obj:
            benji_obj.cleanup(override_lock=override_lock)

    def metadata_export(self, filter_expression: Optional[str], output_file: Optional[str], force: bool) -> None:
        with Benji(self.config) as benji_obj:
            version_uid_objs = [version.uid for version in benji_obj.find_versions_with_filter(filter_expression)]
            if output_file is None:
                benji_obj.metadata_export(version_uid_objs, sys.stdout)
            else:
                if os.path.exists(output_file) and not force:
                    raise FileExistsError('The output file already exists.')

                with open(output_file, 'w') as f:
                    benji_obj.metadata_export(version_uid_objs, f)

    def metadata_backup(self, filter_expression: str, force: bool = False) -> None:
        with Benji(self.config) as benji_obj:
            version_uid_objs = [version.uid for version in benji_obj.find_versions_with_filter(filter_expression)]
            benji_obj.metadata_backup(version_uid_objs, overwrite=force)

    def metadata_import(self, input_file: str = None) -> None:
        with Benji(self.config) as benji_obj:
            if input_file is None:
                benji_obj.metadata_import(sys.stdin)
            else:
                with open(input_file, 'r') as f:
                    benji_obj.metadata_import(f)

    def metadata_restore(self, version_uids: List[str], storage: str = None) -> None:
        version_uid_objs = [VersionUid(version_uid) for version_uid in version_uids]
        with Benji(self.config) as benji_obj:
            benji_obj.metadata_restore(version_uid_objs, storage)

    @staticmethod
    def _metadata_ls_table_output(version_uids: List[VersionUid]):
        tbl = PrettyTable()
        tbl.field_names = ['uid']
        tbl.align['uid'] = 'l'
        for version_uid in version_uids:
            tbl.add_row([version_uid])
        print(tbl)

    def metadata_ls(self, storage: str = None) -> None:
        with Benji(self.config) as benji_obj:
            version_uids = benji_obj.metadata_ls(storage)
            if self.machine_output:
                json.dump(
                    [version_uid for version_uid in version_uids],
                    sys.stdout,
                    indent=2,
                )
            else:
                self._metadata_ls_table_output(version_uids)

    def label(self, version_uid: str, labels: List[str]) -> None:
        version_uid_obj = VersionUid(version_uid)
        label_add, label_remove = InputValidation.parse_and_validate_labels(labels)
        with Benji(self.config) as benji_obj:
            for name, value in label_add:
                benji_obj.add_label(version_uid_obj, name, value)
            for name in label_remove:
                benji_obj.rm_label(version_uid_obj, name)
            if label_add:
                logger.info('Added label(s) to version {}: {}.'.format(
                    version_uid_obj, ', '.join('{}={}'.format(name, value) for name, value in label_add)))
            if label_remove:
                logger.info('Removed label(s) from version {}: {}.'.format(version_uid_obj, ', '.join(label_remove)))

    def database_init(self) -> None:
        Benji(self.config, init_database=True).close()

    def database_migrate(self) -> None:
        Benji(self.config, migrate_database=True).close()

    def enforce_retention_policy(self, rules_spec: str, filter_expression: str, dry_run: bool,
                                 keep_metadata_backup: bool, group_label: Optional[str]) -> None:
        with Benji(self.config) as benji_obj:
            dismissed_versions = benji_obj.enforce_retention_policy(filter_expression=filter_expression,
                                                                    rules_spec=rules_spec,
                                                                    dry_run=dry_run,
                                                                    keep_metadata_backup=keep_metadata_backup,
                                                                    group_label=group_label)
            if self.machine_output:
                benji_obj.export_any({
                    'versions': dismissed_versions,
                },
                                     sys.stdout,
                                     ignore_relationships=(((Version,), ('blocks',)),))

    def nbd(self, bind_address: str, bind_port: str, read_only: bool) -> None:
        with Benji(self.config) as benji_obj:
            store = BenjiStore(benji_obj)
            addr = (bind_address, bind_port)
            server = NbdServer(addr, store, read_only)
            logger.info("Starting to serve NBD on %s:%s" % (addr[0], addr[1]))
            server.serve_forever()

    def version_info(self) -> None:
        if not self.machine_output:
            logger.info('Benji version: {}.'.format(__version__))
            logger.info('Configuration version: {}, supported {}.'.format(VERSIONS.configuration.current,
                                                                          VERSIONS.configuration.supported))
            logger.info('Metadata version: {}, supported {}.'.format(VERSIONS.database_metadata.current,
                                                                     VERSIONS.database_metadata.supported))
            logger.info('Object metadata version: {}, supported {}.'.format(VERSIONS.object_metadata.current,
                                                                            VERSIONS.object_metadata.supported))
        else:
            result = {
                'version': __version__,
                'configuration_version': {
                    'current': str(VERSIONS.configuration.current),
                    'supported': str(VERSIONS.configuration.supported)
                },
                'database_metadata_version': {
                    'current': str(VERSIONS.database_metadata.current),
                    'supported': str(VERSIONS.database_metadata.supported)
                },
                'object_metadata_version': {
                    'current': str(VERSIONS.object_metadata.current),
                    'supported': str(VERSIONS.object_metadata.supported)
                },
            }
            print(json.dumps(result, indent=4))

    @staticmethod
    def _ls_storage_stats_table_output(objects_count: int, objects_size: int) -> None:
        tbl = PrettyTable()
        tbl.field_names = [
            'objects_count',
            'objects_size',
        ]
        tbl.align['objects_count'] = 'r'
        tbl.align['objects_size'] = 'r'
        row = [
            objects_count,
            PrettyPrint.bytes(objects_size),
        ]
        tbl.add_row(row)
        print(tbl)

    def storage_stats(self, storage_name: str = None) -> None:
        with Benji(self.config) as benji_obj:
            objects_count, objects_size = benji_obj.storage_stats(storage_name)

            if self.machine_output:
                result = {
                    'objects_count': objects_count,
                    'objects_size': objects_size,
                }
                print(json.dumps(result, indent=4))
            else:
                self._ls_storage_stats_table_output(objects_count, objects_size)

    @staticmethod
    def _storage_usage_table_output(usage: Dict[str, Dict[str, int]]) -> None:
        tbl = PrettyTable()
        tbl.field_names = [
            'storage',
            'virtual',
            'sparse',
            'shared',
            'exclusive',
            'deduplicated_exclusive',
        ]
        tbl.align['storage'] = 'l'
        tbl.align['virtual'] = 'l'
        tbl.align['sparse'] = 'l'
        tbl.align['shared'] = 'r'
        tbl.align['exclusive'] = 'r'
        tbl.align['deduplicated_exclusive'] = 'r'
        for storage_name, usage in usage.items():
            row = [
                storage_name,
                PrettyPrint.bytes(usage['virtual']),
                PrettyPrint.bytes(usage['sparse']),
                PrettyPrint.bytes(usage['shared']),
                PrettyPrint.bytes(usage['exclusive']),
                PrettyPrint.bytes(usage['deduplicated_exclusive']),
            ]
            tbl.add_row(row)
        print(tbl)

    def storage_usage(self, filter_expression: str):
        with Benji(self.config) as benji_obj:
            usage = benji_obj.storage_usage(filter_expression)
            if self.machine_output:
                print(json.dumps(usage, indent=4))
            else:
                self._storage_usage_table_output(usage)

    def rest_api(self, bind_address: str, bind_port: int, threads: int) -> None:
        from benji.restapi import RestAPI
        api = RestAPI(self.config)
        logger.info(f'Starting REST API via gunicorn on {bind_address}:{bind_port}.')
        debug = bool(logger.isEnabledFor(logging.DEBUG))
        api.run(bind_address=bind_address, bind_port=bind_port, threads=threads, debug=debug)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# PYTHON_ARGCOMPLETE_OK

import argparse
import os
import sys
from functools import partial
from typing import NamedTuple, Type, Optional

import argcomplete

from benji.exception import InternalError
from benji.io.factory import IOFactory
from benji.storage.factory import StorageFactory


class _ExceptionMapping(NamedTuple):
    exception: Type[BaseException]
    exit_code: int
    include_stacktrace: bool


def completion(shell: str) -> None:
    print(argcomplete.shellcode(sys.argv[0], shell=shell))


def integer_range(minimum: int, maximum: int, arg: str) -> Optional[int]:
    if arg is None:
        return None

    try:
        value = int(arg)
    except ValueError as err:
        raise argparse.ArgumentTypeError(str(err))

    if value < minimum or (maximum is not None and value > maximum):
        raise argparse.ArgumentTypeError('Expected a value between {} and {}, got {}.'.format(minimum, maximum, value))

    return value


def main():
    if sys.hexversion < 0x030605F0:
        # We're using features introduced with Python 3.6. In addition Python versions before 3.6.5 have some
        # shortcomings in the concurrent.futures implementation which lead to an excessive memory usage.
        raise InternalError('Benji only supports Python 3.6.5 or above.')

    enable_experimental = os.getenv('BENJI_EXPERIMENTAL', default='0') == '1'

    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter, allow_abbrev=False)

    parser.add_argument('-c', '--config-file', default=None, type=str, help='Specify a non-default configuration file')
    parser.add_argument('-m',
                        '--machine-output',
                        action='store_true',
                        default=False,
                        help='Enable machine-readable JSON output')
    parser.add_argument('--log-level',
                        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR'],
                        default='INFO',
                        help='Only log messages of this level or above on the console')
    parser.add_argument('--no-color',
                        action='store_true',
                        default=False,
                        help='Disable colorization of console logging')

    subparsers_root = parser.add_subparsers(title='commands')

    # BACKUP
    p = subparsers_root.add_parser('backup', help='Perform a backup')
    p.add_argument('-u',
                   '--uid',
                   dest='version_uid',
                   default=None,
                   help='Unique ID of created version (will be generated automatically if not specified)')
    p.add_argument('-s', '--snapshot', default='', help='Snapshot name (e.g. the name of the RBD snapshot)')
    p.add_argument('-r', '--rbd-hints', default=None, help='Hints in rbd diff JSON format')
    p.add_argument('-f', '--base-version', dest='base_version_uid', default=None, help='Base version UID')
    p.add_argument('-b', '--block-size', type=int, default=None, help='Block size in bytes')
    p.add_argument('-l',
                   '--label',
                   action='append',
                   dest='labels',
                   metavar='label',
                   default=None,
                   help='Labels for this version (can be repeated)')
    p.add_argument('-S', '--storage', default='', help='Destination storage (if unspecified the default is used)')
    p.add_argument('source', help='Source URL')
    p.add_argument('volume', help='Volume name')
    p.set_defaults(func='backup')

    # BATCH-DEEP-SCRUB
    p = subparsers_root.add_parser('batch-deep-scrub',
                                   help='Check data and metadata integrity of multiple versions at once',
                                   formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    p.add_argument('-p',
                   '--block-percentage',
                   type=partial(integer_range, 1, 100),
                   default=100,
                   help='Check only a certain percentage of blocks')
    p.add_argument('-P',
                   '--version-percentage',
                   type=partial(integer_range, 1, 100),
                   default=100,
                   help='Check only a certain percentage of versions')
    p.add_argument('-g', '--group_label', default=None, help='Label to find related versions')
    p.add_argument('filter_expression', nargs='?', default=None, help='Version filter expression')
    p.set_defaults(func='batch_deep_scrub')

    # BATCH-SCRUB
    p = subparsers_root.add_parser('batch-scrub',
                                   help='Check block existence and metadata integrity of multiple versions at once',
                                   formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    p.add_argument('-p',
                   '--block-percentage',
                   type=partial(integer_range, 1, 100),
                   default=100,
                   help='Check only a certain percentage of blocks')
    p.add_argument('-P',
                   '--version-percentage',
                   type=partial(integer_range, 1, 100),
                   default=100,
                   help='Check only a certain percentage of versions')
    p.add_argument('-g', '--group_label', default=None, help='Label to find related versions')
    p.add_argument('filter_expression', nargs='?', default=None, help='Version filter expression')
    p.set_defaults(func='batch_scrub')

    # CLEANUP
    p = subparsers_root.add_parser('cleanup', help='Cleanup no longer referenced blocks')
    p.add_argument('--override-lock', action='store_true', help='Override and release any held lock (dangerous)')
    p.set_defaults(func='cleanup')

    # COMPLETION
    p = subparsers_root.add_parser('completion', help='Emit autocompletion script')
    p.add_argument('shell', choices=['bash', 'tcsh'], help='Shell')
    p.set_defaults(func='completion')

    # DATABASE-INIT
    p = subparsers_root.add_parser('database-init',
                                   help='Initialize the database (will not delete existing tables or data)')
    p.set_defaults(func='database_init')

    # DATABASE-MIGRATE
    p = subparsers_root.add_parser('database-migrate', help='Migrate an existing database to a new schema revision')
    p.set_defaults(func='database_migrate')

    # DEEP-SCRUB
    p = subparsers_root.add_parser('deep-scrub',
                                   help='Check a version\'s data and metadata integrity',
                                   formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    p.add_argument('-s', '--source', default=None, help='Additionally compare version against source URL')
    p.add_argument('-p',
                   '--block-percentage',
                   type=partial(integer_range, 1, 100),
                   default=100,
                   help='Check only a certain percentage of blocks')
    p.add_argument('version_uid', help='Version UID')
    p.set_defaults(func='deep_scrub')

    # ENFORCE
    p = subparsers_root.add_parser('enforce', help="Enforce a retention policy ")
    p.add_argument('--dry-run', action='store_true', help='Only show which versions would be removed')
    p.add_argument('-k', '--keep-metadata-backup', action='store_true', help='Keep version metadata backup')
    p.add_argument('-g', '--group_label', default=None, help='Label to find related versions to remove')
    p.add_argument('rules_spec', help='Retention rules specification')
    p.add_argument('filter_expression', nargs='?', default=None, help='Version filter expression')
    p.set_defaults(func='enforce_retention_policy')

    # LABEL
    p = subparsers_root.add_parser('label', help='Add labels to a version')
    p.add_argument('version_uid')
    p.add_argument('labels', nargs='+')
    p.set_defaults(func='label')

    # LS
    p = subparsers_root.add_parser('ls', help='List versions')
    p.add_argument('filter_expression', nargs='?', default=None, help='Version filter expression')
    p.add_argument('-l', '--include-labels', action='store_true', help='Include labels in output')
    p.add_argument('-s', '--include-stats', action='store_true', help='Include statistics in output')
    p.set_defaults(func='ls')

    # METADATA-BACKUP
    p = subparsers_root.add_parser('metadata-backup', help='Back up the metadata of one or more versions')
    p.add_argument('filter_expression', help="Version filter expression")
    p.add_argument('-f', '--force', action='store_true', help='Overwrite existing metadata backups')
    p.set_defaults(func='metadata_backup')

    # METADATA EXPORT
    p = subparsers_root.add_parser('metadata-export',
                                   help='Export the metadata of one or more versions to a file or standard output')
    p.add_argument('filter_expression', nargs='?', default=None, help="Version filter expression")
    p.add_argument('-f', '--force', action='store_true', help='Overwrite an existing output file')
    p.add_argument('-o', '--output-file', default=None, help='Output file (standard output if missing)')
    p.set_defaults(func='metadata_export')

    # METADATA-IMPORT
    p = subparsers_root.add_parser('metadata-import',
                                   help='Import the metadata of one or more versions from a file or standard input')
    p.add_argument('-i', '--input-file', default=None, help='Input file (standard input if missing)')
    p.set_defaults(func='metadata_import')

    # METADATA-LS
    p = subparsers_root.add_parser('metadata-ls', help='List the version metadata backup')
    p.add_argument('-S', '--storage', default=None, help='Source storage (if unspecified the default is used)')
    p.set_defaults(func='metadata_ls')

    # METADATA-RESTORE
    p = subparsers_root.add_parser('metadata-restore', help='Restore the metadata of one ore more versions')
    p.add_argument('-S', '--storage', default=None, help='Source storage (if unspecified the default is used)')
    p.add_argument('version_uids', metavar='VERSION_UID', nargs='+', help="Version UID")
    p.set_defaults(func='metadata_restore')

    # NBD
    p = subparsers_root.add_parser('nbd',
                                   help='Start an NBD server',
                                   formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    p.add_argument('-a', '--bind-address', default='127.0.0.1', help='Bind to the specified IP address')
    p.add_argument('-p', '--bind-port', default=10809, help='Bind to the specified port')
    p.add_argument('-r', '--read-only', action='store_true', default=False, help='NBD device is read-only')
    p.set_defaults(func='nbd')

    # PROTECT
    p = subparsers_root.add_parser('protect', help='Protect one or more versions')
    p.add_argument('version_uids', metavar='version_uid', nargs='+', help="Version UID")
    p.set_defaults(func='protect')

    # RESTORE
    p = subparsers_root.add_parser('restore', help='Restore a backup')
    p.add_argument('-s', '--sparse', action='store_true', help='Restore only existing blocks')
    p.add_argument('-f', '--force', action='store_true', help='Overwrite an existing file, device or image')
    p.add_argument('-d', '--database-less', action='store_true', help='Restore without requiring the database')
    p.add_argument('-S', '--storage', default=None, help='Source storage (if unspecified the default is used)')
    p.add_argument('version_uid', help='Version UID to restore')
    p.add_argument('destination', help='Destination URL')
    p.set_defaults(func='restore')

    # RM
    p = subparsers_root.add_parser('rm', help='Remove one or more versions')
    p.add_argument('-f', '--force', action='store_true', help='Force removal (overrides protection of recent versions)')
    p.add_argument('-k', '--keep-metadata-backup', action='store_true', help='Keep version metadata backup')
    p.add_argument('--override-lock', action='store_true', help='Override and release any held locks (dangerous)')
    p.add_argument('version_uids', metavar='version_uid', nargs='+', help='Version UID')
    p.set_defaults(func='rm')

    # SCRUB
    p = subparsers_root.add_parser('scrub',
                                   help='Check a version\'s block existence and metadata integrity',
                                   formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    p.add_argument('-p',
                   '--block-percentage',
                   type=partial(integer_range, 1, 100),
                   default=100,
                   help='Check only a certain percentage of blocks')
    p.add_argument('version_uid', help='Version UID')
    p.set_defaults(func='scrub')

    # STORAGE-STATS
    p = subparsers_root.add_parser('storage-stats', help='Show storage statistics')
    p.add_argument('storage_name', nargs='?', default=None, help='Storage')
    p.set_defaults(func='storage_stats')

    # UNPROTECT
    p = subparsers_root.add_parser('unprotect', help='Unprotect one or more versions')
    p.add_argument('version_uids', metavar='version_uid', nargs='+', help='Version UID')
    p.set_defaults(func='unprotect')

    # VERSION-INFO
    p = subparsers_root.add_parser('version-info', help='Program version information')
    p.set_defaults(func='version_info')

    # REST-API
    if enable_experimental:
        p = subparsers_root.add_parser('rest-api', help='Start REST API server')
        p.set_defaults(func='rest_api')
        p.add_argument('-a', '--bind-address', default='127.0.0.1', help='Bind to the specified IP address')
        p.add_argument('-p', '--bind-port', default=8080, type=int, help='Bind to the specified port')
        p.add_argument('--threads', default=1, type=int, help='Number of worker threads')

    # DU
    p = subparsers_root.add_parser('storage-usage', help='Provide storage usage statistics')
    p.add_argument('filter_expression', nargs='?', default=None, help='Version filter expression')
    p.set_defaults(func='storage_usage')

    argcomplete.autocomplete(parser)
    args = parser.parse_args()

    if not hasattr(args, 'func'):
        parser.print_usage()
        sys.exit(os.EX_USAGE)

    if args.func == 'completion':
        completion(args.shell)
        sys.exit(os.EX_OK)

    from benji.config import Config
    from benji.logging import logger, init_logging
    if args.config_file is not None and args.config_file != '':
        try:
            cfg = open(args.config_file, 'r', encoding='utf-8').read()
        except FileNotFoundError:
            logger.error('File {} not found.'.format(args.config_file))
            sys.exit(os.EX_USAGE)
        config = Config(ad_hoc_config=cfg)
    else:
        config = Config()

    console_formatter = 'console-colored'
    if args.machine_output:
        console_formatter = 'json'
    elif args.no_color:
        console_formatter = 'console-plain'

    init_logging(logfile=config.get('logFile', types=(str, type(None))),
                 console_level=args.log_level,
                 console_formatter=console_formatter)

    IOFactory.initialize(config)
    StorageFactory.initialize(config)

    import benji.commands
    commands = benji.commands.Commands(args.machine_output, config)
    func = getattr(commands, args.func)

    # Pass over to function
    func_args = dict(args._get_kwargs())
    del func_args['config_file']
    del func_args['func']
    del func_args['log_level']
    del func_args['machine_output']
    del func_args['no_color']

    # From most specific to least specific
    # yapf: disable
    exception_mappings = [
        _ExceptionMapping(exception=benji.exception.UsageError, exit_code=os.EX_USAGE, include_stacktrace=False),
        _ExceptionMapping(exception=benji.exception.AlreadyLocked, exit_code=os.EX_NOPERM, include_stacktrace=False),
        _ExceptionMapping(exception=benji.exception.InternalError, exit_code=os.EX_SOFTWARE, include_stacktrace=True),
        _ExceptionMapping(exception=benji.exception.ConfigurationError, exit_code=os.EX_CONFIG, include_stacktrace=False),
        _ExceptionMapping(exception=benji.exception.InputDataError, exit_code=os.EX_DATAERR, include_stacktrace=False),
        _ExceptionMapping(exception=benji.exception.ScrubbingError, exit_code=os.EX_DATAERR, include_stacktrace=False),
        _ExceptionMapping(exception=PermissionError, exit_code=os.EX_NOPERM, include_stacktrace=False),
        _ExceptionMapping(exception=FileExistsError, exit_code=os.EX_CANTCREAT, include_stacktrace=False),
        _ExceptionMapping(exception=FileNotFoundError, exit_code=os.EX_NOINPUT, include_stacktrace=False),
        _ExceptionMapping(exception=EOFError, exit_code=os.EX_IOERR, include_stacktrace=True),
        _ExceptionMapping(exception=IOError, exit_code=os.EX_IOERR, include_stacktrace=True),
        _ExceptionMapping(exception=OSError, exit_code=os.EX_OSERR, include_stacktrace=True),
        _ExceptionMapping(exception=ConnectionError, exit_code=os.EX_IOERR, include_stacktrace=True),
        _ExceptionMapping(exception=LookupError, exit_code=os.EX_NOINPUT, include_stacktrace=True),
        _ExceptionMapping(exception=KeyboardInterrupt, exit_code=os.EX_NOINPUT, include_stacktrace=False),
        _ExceptionMapping(exception=BaseException, exit_code=os.EX_SOFTWARE, include_stacktrace=True),
    ]
    # yapf: enable

    try:
        logger.debug('commands.{0}(**{1!r})'.format(args.func, func_args))
        func(**func_args)
        sys.exit(os.EX_OK)
    except SystemExit:
        raise
    except BaseException as exception:
        for case in exception_mappings:
            if isinstance(exception, case.exception):
                message = str(exception)
                if message:
                    message = '{}: {}'.format(exception.__class__.__name__, message)
                else:
                    message = '{} exception occurred.'.format(exception.__class__.__name__)
                if case.include_stacktrace:
                    logger.error(message, exc_info=True)
                else:
                    logger.debug(message, exc_info=True)
                    logger.error(message)
                sys.exit(case.exit_code)


if __name__ == '__main__':
    main()



import logging
import os
import random
import shutil
import string
from binascii import hexlify

from benji.benji import Benji
from benji.config import Config
from benji.database import Database
from benji.io.factory import IOFactory
from benji.logging import init_logging
from benji.storage.factory import StorageFactory


class _TestPath:

    def __init__(self):
        self.path = '../../../tests-scratch/benji-test_' + TestCaseBase.random_string(16)
        for dir in [
                self.path,
                self.path + '/data',
                self.path + '/data-2',
                self.path + '/lock',
        ]:
            os.makedirs(dir, exist_ok=True)

    def close(self):
        shutil.rmtree(self.path)


class TestCaseBase:

    @staticmethod
    def random_string(length):
        return ''.join(random.choices(string.ascii_letters + string.digits, k=length))

    @staticmethod
    def random_bytes(length):
        return bytes(random.getrandbits(8) for _ in range(length))

    @staticmethod
    def random_hex(length):
        return hexlify(bytes(random.getrandbits(8) for _ in range(length))).decode('ascii')

    def setUp(self):
        self.testpath = _TestPath()
        init_logging(console_level=logging.WARN if os.environ.get('UNITTEST_QUIET', False) else logging.DEBUG,
                     console_formatter='console-plain')
        self.config = Config(ad_hoc_config=self.CONFIG.format(testpath=self.testpath.path))

    def tearDown(self):
        self.testpath.close()


class StorageTestCaseBase(TestCaseBase):

    def setUp(self):
        super().setUp()

        StorageFactory.initialize(self.config)

        default_storage_name = self.config.get('defaultStorage', types=str)
        self.storage = StorageFactory.get_by_name(default_storage_name)
        for block_uid in self.storage.list_blocks():
            self.storage.rm_block(block_uid)
        for version_uid in self.storage.list_versions():
            self.storage.rm_version(version_uid)

    def tearDown(self):
        uids = list(self.storage.list_blocks())
        self.assertEqual(0, len(uids))
        StorageFactory.close()
        super().tearDown()


class DatabaseBackendTestCaseBase(TestCaseBase):

    def setUp(self):
        super().setUp()

        Database.configure(self.config)
        Database.init(_destroy=True)
        Database.open()

    def tearDown(self):
        Database.close()
        super().tearDown()


class BenjiTestCaseBase(TestCaseBase):

    def setUp(self):
        super().setUp()
        IOFactory.initialize(self.config)
        StorageFactory.initialize(self.config)

    def tearDown(self):
        StorageFactory.close()
        super().tearDown()

    def benji_open(self, init_database=False, in_memory_database=False):
        self.benji = Benji(self.config,
                           init_database=init_database,
                           in_memory_database=in_memory_database,
                           _destroy_database=init_database)
        return self.benji



from setuptools import setup, find_packages

setup(name='benji-k8s-tools',
      version='0.1',
      description='Small tools for using Benji with Kubernetes',
      url='https://github.com/elemental-lf/benji',
      author='Lars Fenneberg',
      author_email='lf@elemental.net',
      license='LGPG-3',
      python_requires='~=3.6',
      packages=find_packages('src'),
      package_dir={
          '': 'src',
      },
      install_requires=['benji', 'kubernetes>=10.0.0,<11'],
      entry_points="""
        [console_scripts]
            benji-backup-pvc = benji.k8s_tools.scripts.backup_pvc:main
            benji-command = benji.k8s_tools.scripts.command:main
            benji-restore-pvc = benji.k8s_tools.scripts.restore_pvc:main
            benji-versions-status = benji.k8s_tools.scripts.versions_status:main
    """)



#
# This is loosely based on timegaps/timefilter.py and timegaps/main.py
# which are part of the timegaps module. Please see: https://github.com/jgehrcke/timegaps.
#
# The original copyright and license are:
#
# Copyright 2014 Jan-Philip Gehrcke (http://gehrcke.de)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import datetime
import re
from collections import OrderedDict
from collections import defaultdict
from typing import List, Dict, Sequence, Union, Tuple, Set

import dateutil

from benji.database import Version
from benji.exception import UsageError
from benji.logging import logger
from benji.repr import ReprMixIn


class RetentionFilter(ReprMixIn):

    _valid_categories = ('latest', 'hours', 'days', 'weeks', 'months', 'years')

    # This method is taken from timegaps/main.py, its original name is parse_rules_from_cmdline.
    @classmethod
    def _parse_rules(cls, rules_spec: str) -> OrderedDict:
        tokens = rules_spec.split(',')
        rules_dict: Dict[str, int] = {}
        for token in tokens:
            if not token:
                raise ValueError('Empty retention policy element.')
            match = re.search(r'^([a-z]+)([0-9]+)$', token)
            if match:
                category = match.group(1)
                timecount = int(match.group(2))
                if category not in cls._valid_categories:
                    raise ValueError('Time category {} in retention policy is invalid.'.format(category))
                if category in rules_dict:
                    raise ValueError('Time category {} listed more than once in retention policy.'.format(category))
                if timecount <= 0:
                    raise UsageError('Count of time category {} must be a positive integer.'.format(category))
                rules_dict[category] = timecount
                continue
            raise ValueError('Invalid retention policy element {}.'.format(token))

        rules: OrderedDict[str, int] = OrderedDict()
        for category in cls._valid_categories:
            if category in rules_dict:
                rules[category] = rules_dict[category]

        return rules

    def __init__(self, rules_spec: str, reference_time: datetime.datetime = None, tz: datetime.tzinfo = None) -> None:
        self.tz = tz if tz is not None else dateutil.tz.tzlocal()
        if reference_time is None:
            self.reference_time = datetime.datetime.now(tz=self.tz)
        else:
            if reference_time.tzinfo is None:
                # Assume it is in UTC
                self.reference_time = reference_time.replace(tzinfo=datetime.timezone.utc)
            else:
                self.reference_time = reference_time

        self.rules = self._parse_rules(rules_spec)

        logger.debug('Retention filter set up with reference time {} and rules {}.'.format(
            self.reference_time.isoformat(timespec='seconds'), self.rules))

    def filter(self, versions: Union[Sequence[Version], Set[Version]]) -> List[Version]:
        return self._filter(versions)[0]

    def _filter(self, versions: Union[Sequence[Version], Set[Version]]
               ) -> Tuple[List[Version], Dict[str, Dict[int, List[Version]]]]:
        # Category labels without latest
        categories = [category for category in self.rules.keys() if category != 'latest']

        versions_by_category: Dict[str, Dict[int, List[Version]]] = {}
        versions_by_category_remaining: Dict[str, Dict[int, List[Version]]] = {}
        for category in categories:
            versions_by_category[category] = defaultdict(list)
            versions_by_category_remaining[category] = {}

        # Make our own copy
        versions = list(versions)
        # Sort from youngest to oldest
        versions.sort(key=lambda version: version.date, reverse=True)

        # Remove latest versions from consideration if configured
        if 'latest' in self.rules:
            logger.debug('Keeping {} latest versions.'.format(self.rules['latest']))
            versions_by_category_remaining['latest'] = {0: versions[:self.rules['latest']]}
            del versions[:self.rules['latest']]

        dismissed_versions = []
        for version in versions:
            try:
                # version.date is naive and in UTC, attach time zone to make it time zone aware.
                td = _Timedelta(version.date.replace(tzinfo=datetime.timezone.utc), self.reference_time, tz=self.tz)
            except ValueError as exception:
                # Err on the safe side, ignore this versions (i.e. it won't be dismissed)
                logger.warning('Version {}: {}.'.format(version.uid, exception))
                continue

            logger.debug('Time and time delta for version {} are {} and {}.'.format(
                version.uid, version.date.isoformat(timespec='seconds'), td))

            for category in categories:
                timecount = getattr(td, category)
                if timecount <= self.rules[category]:
                    logger.debug('Found matching category {}, timecount {}.'.format(category, timecount))
                    versions_by_category[category][timecount].append(version)
                    break
            else:
                # For loop did not break: The item doesn't fit into any category, it's too old.
                dismissed_versions.append(version)
                logger.debug('Dismissing version, it doesn\'t fit into any category.')

        for category in categories:
            for timecount in versions_by_category[category]:
                # Keep the oldest of each category, reject the rest
                dismissed_versions.extend(versions_by_category[category][timecount][:-1])
                versions_by_category_remaining[category][timecount] = versions_by_category[category][timecount][-1:]

        return dismissed_versions, versions_by_category_remaining


class _Timedelta(ReprMixIn):

    @staticmethod
    def _round_down(t: datetime.datetime, *, start_of: str):
        if start_of == 'hour':
            return t.replace(minute=0, second=0, microsecond=0)
        elif start_of == 'day':
            return t.replace(hour=0, minute=0, second=0, microsecond=0)
        elif start_of == 'week':
            # This will round down to the last Monday at 00:00.0 before t.
            return t + dateutil.relativedelta.relativedelta(
                weekday=dateutil.relativedelta.MO(-1), hour=0, minute=0, second=0, microsecond=0)
        elif start_of == 'month':
            return t.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
        elif start_of == 'year':
            return t.replace(month=1, day=1, hour=0, minute=0, second=0, microsecond=0)
        else:
            raise ValueError('Start of {} is unknown.'.format(start_of))

    def __init__(self, t: datetime.datetime, reference_time: datetime.datetime, tz: datetime.tzinfo) -> None:
        if t.tzinfo is None:
            raise ValueError('Time is not time zone aware.')
        if reference_time.tzinfo is None:
            raise ValueError('Reference time is not time zone aware.')
        if reference_time - t < datetime.timedelta(0):
            raise ValueError('{} isn\'t earlier than the reference time {} (difference = {}).'.format(
                t.isoformat(timespec='seconds'), reference_time.isoformat(timespec='seconds'),
                (reference_time - t).total_seconds()))

        t = t.astimezone(tz=tz)
        reference_time = reference_time.astimezone(tz=tz)

        # Hours
        delta = self._round_down(reference_time, start_of='hour') - self._round_down(t, start_of='hour')
        self.hours = int(delta.total_seconds() // 3600)

        # Days
        delta = self._round_down(reference_time, start_of='day') - self._round_down(t, start_of='day')
        self.days = int(delta.days)

        # Weeks
        delta = self._round_down(reference_time, start_of='week') - self._round_down(t, start_of='week')
        self.weeks = int(delta // datetime.timedelta(weeks=1))

        # Months
        delta = dateutil.relativedelta.relativedelta(self._round_down(reference_time, start_of='month'),
                                                     self._round_down(t, start_of='month')).normalized()
        self.months = delta.years * 12 + delta.months

        # Years
        delta = dateutil.relativedelta.relativedelta(self._round_down(reference_time, start_of='year'),
                                                     self._round_down(t, start_of='year')).normalized()
        self.years = delta.years



import random

from benji.blockuidhistory import BlockUidHistory
from benji.database import BlockUid


def test_seen():
    history = BlockUidHistory()
    blocks_in_1 = set()
    blocks_out_1 = set()
    block_exists_1 = set()
    blocks_in_2 = set()
    blocks_out_2 = set()
    block_exists_2 = set()
    for i in range(0, 100000):
        block = BlockUid(random.randint(1, 2**8), random.randint(1, 2**64 - 1))
        if block in block_exists_1:
            continue
        block_exists_1.add(block)
        if random.randint(1, 100) > 20:
            blocks_in_1.add(block)
            history.add(1, block)
            assert history.seen(1, block)
        else:
            blocks_out_1.add(block)
            assert not history.seen(1, block)
    for i in range(0, 100000):
        block = BlockUid(random.randint(1, 2**8), random.randint(1, 2**20))
        if block in block_exists_2:
            continue
        block_exists_2.add(block)
        if random.randint(1, 100) > 20:
            blocks_in_2.add(block)
            history.add(2, block)
            assert history.seen(2, block)
        else:
            blocks_out_2.add(block)
            assert not history.seen(2, block)
    for block in blocks_in_1:
        assert history.seen(1, block)
    for block in blocks_out_1:
        assert not history.seen(1, block)
    for block in blocks_in_2:
        assert history.seen(2, block)
    for block in blocks_out_2:
        assert not history.seen(2, block)


def test_oom():
    history = BlockUidHistory()
    for i in range(0, 1000000):
        storage_id = random.randint(1, 4)
        block = BlockUid(random.randint(1, 2**8), random.randint(1, 2**24))
        history.add(storage_id, block)
        assert history.seen(storage_id, block)



"""Add version status

Revision ID: b1fa564a0ebf
Revises: 2e028f08346b
Create Date: 2019-01-21 19:25:42.193956

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = 'b1fa564a0ebf'
down_revision = '2e028f08346b'
branch_labels = None
depends_on = None

versions = sa.table('versions', sa.column('valid', sa.Boolean), sa.column('status', sa.Integer))


def upgrade():
    with op.batch_alter_table('versions', schema=None) as batch_op:
        batch_op.add_column(sa.Column('status', sa.Integer, nullable=False, server_default=sa.text('2')))
        batch_op.create_check_constraint('status', 'status >= 1 AND status <= 3')

    op.execute(versions.update().where(versions.c.valid == op.inline_literal(False, type_=sa.Boolean)).values({
        'status':
        op.inline_literal(3)
    }))

    with op.batch_alter_table('versions', schema=None) as batch_op:
        batch_op.drop_column('valid')


def downgrade():
    pass



#!/usr/bin/env python3
import argparse
import logging
import random
import string
import sys
import time
from typing import Any, Dict, Optional, Tuple

import kubernetes
import kubernetes.stream

import benji.helpers.ceph as ceph
import benji.helpers.prometheus as prometheus
import benji.helpers.settings as settings
import benji.helpers.utils as utils
import benji.k8s_tools.kubernetes

FSFREEZE_TIMEOUT = 15
FSFREEZE_UNFREEZE_TRIES = (0, 1, 1, 1, 15, 30)
FSFREEZE_ANNOTATION = 'benji-backup.me/fsfreeze'
FSFREEZE_POD_LABEL_SELECTOR = 'benji-backup.me/component=fsfreeze'
FSFREEZE_CONTAINER_NAME = 'fsfreeze'
KUBELET_MOUNT_PATH_FORMAT = '/var/lib/kubelet/plugins/kubernetes.io/rbd/mounts/rbd-image-{}'

utils.setup_logging()
logger = logging.getLogger()


def _random_string(length: int, characters: str = string.ascii_lowercase + string.digits) -> str:
    return ''.join(random.choice(characters) for _ in range(length))


def _determine_fsfreeze_info(pvc_namespace: str, pvc_name: str,
                             image: str) -> Tuple[bool, Optional[str], Optional[str], str]:
    pv_fsfreeze = False
    pv_host_ip = None
    pv_fsfreeze_pod = None
    pv_mount_point = KUBELET_MOUNT_PATH_FORMAT.format(image)

    core_v1_api = kubernetes.client.CoreV1Api()
    pvc = core_v1_api.read_namespaced_persistent_volume_claim(pvc_name, pvc_namespace)
    service_account_namespace = benji.k8s_tools.kubernetes.service_account_namespace()
    if hasattr(pvc.metadata,
               'annotations') and FSFREEZE_ANNOTATION in pvc.metadata.annotations and pvc.metadata.annotations[FSFREEZE_ANNOTATION] == 'yes':
        pods = core_v1_api.list_namespaced_pod(service_account_namespace, watch=False).items
        for pod in pods:
            if pv_fsfreeze:
                break
            if not hasattr(pod.spec, 'volumes'):
                continue
            for volume in pod.spec.volumes:
                if not hasattr(volume, 'persistent_volume_claim') or not hasattr(
                        volume.persistent_volume_claim, 'claim_name') or volume.persistent_volume_claim.claim_name != pvc_name:
                    continue
                if hasattr(pod.status, 'host_ip') and pod.status.host_ip != '':
                    pv_fsfreeze = True
                    pv_host_ip = pod.status.host_ip
                break

        if pv_fsfreeze:
            pods = core_v1_api.list_namespaced_pod(benji.k8s_tools.kubernetes.service_account_namespace(),
                                                   label_selector=FSFREEZE_POD_LABEL_SELECTOR).items

            if not pods:
                logger.error('No fsfreeze pods found (label selector {FSFREEZE_POD_LABEL_SELECTOR}).')

            for pod in pods:
                if not hasattr(pod.status, 'host_ip') or not hasattr(pod.status, 'phase'):
                    continue

                if pod.status.host_ip == pv_host_ip and pod.status.phase == 'Running':
                    pv_fsfreeze_pod = pod.metadata.name
                    break
            else:
                pv_fsfreeze = False

    return pv_fsfreeze, pv_host_ip, pv_fsfreeze_pod, pv_mount_point


@ceph.signal_snapshot_create_pre.connect
def ceph_snapshot_create_pre(sender: str, volume: str, pool: str, image: str, snapshot: str,
                             context: Dict[str, Any]) -> None:
    assert isinstance(context, dict)
    assert 'pvc' in context
    pvc_namespace = context['pvc'].metadata.namespace
    pvc_name = context['pvc'].metadata.name

    pv_fsfreeze, pv_host_ip, pv_fsfreeze_pod, pv_mount_point = _determine_fsfreeze_info(pvc_namespace, pvc_name, image)

    # Record for use in post signals
    context['pv-fsfreeze'] = pv_fsfreeze
    context['pv-host-ip'] = pv_host_ip
    context['pv-fsfreeze-pod'] = pv_fsfreeze_pod
    context['pv-mount-point'] = pv_mount_point

    if not pv_fsfreeze:
        return
    if pv_host_ip is None:
        logger.info(f'PV is not mounted anywhere, skipping fsfreeze.')
        return
    if pv_fsfreeze_pod is None:
        logger.info(f'No fsfreeze pod found for host {pv_host_ip}, skipping fsfreeze for this PV.')
        return

    logger.info(f'Freezing filesystem {pv_mount_point} on host {pv_host_ip} (pod {pv_fsfreeze_pod}).')

    service_account_namespace = benji.k8s_tools.kubernetes.service_account_namespace()
    try:
        benji.k8s_tools.kubernetes.pod_exec(['fsfreeze', '--freeze', pv_mount_point],
                                            name=pv_fsfreeze_pod,
                                            namespace=service_account_namespace,
                                            container=FSFREEZE_CONTAINER_NAME,
                                            timeout=FSFREEZE_TIMEOUT)
    except Exception as exception:
        # Try to unfreeze in any case
        try:
            benji.k8s_tools.kubernetes.pod_exec(['fsfreeze', '--unfreeze', pv_mount_point],
                                                name=pv_fsfreeze_pod,
                                                namespace=service_account_namespace,
                                                container=FSFREEZE_CONTAINER_NAME,
                                                timeout=FSFREEZE_TIMEOUT)
        except Exception as exception_2:
            raise exception_2 from exception
        else:
            raise exception

    logger.debug(f'Freezing filesystem succeeded.')


@ceph.signal_snapshot_create_post_success.connect
def ceph_snapshot_create_post_success(sender: str, volume: str, pool: str, image: str, snapshot: str,
                                      context: Dict[str, Any]) -> None:
    assert isinstance(context, dict)
    pv_fsfreeze = context['pv-fsfreeze']
    if not pv_fsfreeze:
        return

    pv_host_ip = context['pv-host-ip']
    pv_fsfreeze_pod = context['pv-fsfreeze-pod']
    pv_mount_point = context['pv-mount-point']

    logger.info(f'Unfreezing filesystem {pv_mount_point} on host {pv_host_ip}.')

    service_account_namespace = benji.k8s_tools.kubernetes.service_account_namespace()
    for delay in FSFREEZE_UNFREEZE_TRIES:
        if delay > 0:
            time.sleep(delay)

        try:
            benji.k8s_tools.kubernetes.pod_exec(['fsfreeze', '--unfreeze', pv_mount_point],
                                                name=pv_fsfreeze_pod,
                                                namespace=service_account_namespace,
                                                container=FSFREEZE_CONTAINER_NAME,
                                                timeout=FSFREEZE_TIMEOUT)
        except Exception:
            pass
        else:
            logger.debug(f'Unfreezing filesystem succeeded.')
            break
    else:
        logger.error(f'Giving up on unfreezing filesystem {pv_mount_point} on host {pv_host_ip}.')


@ceph.signal_snapshot_create_post_error.connect
def ceph_snapshot_create_post_error(sender: str, volume: str, pool: str, image: str, snapshot: str,
                                    context: Dict[str, Any], exception: Exception) -> None:
    ceph_snapshot_create_post_success(sender, volume, pool, image, snapshot, context)
    raise exception


@ceph.signal_backup_pre.connect
def ceph_backup_pre(sender: str, volume: str, pool: str, image: str, version_labels: Dict[str, str],
                    context: Dict[str, Any]):
    assert isinstance(context, dict)
    context['backup-start-time'] = start_time = time.time()
    prometheus.backup_start_time.labels(volume=volume).set(start_time)


def _k8s_create_pvc_event(type: str, reason: str, message: str, context: Dict[str, Any]):
    assert isinstance(context, dict)
    assert 'pvc' in context
    pvc_namespace = context['pvc'].metadata.namespace
    pvc_name = context['pvc'].metadata.name
    pvc_uid = context['pvc'].metadata.uid

    try:
        benji.k8s_tools.kubernetes.create_pvc_event(type=type,
                                                    reason=reason,
                                                    message=message,
                                                    pvc_namespace=pvc_namespace,
                                                    pvc_name=pvc_name,
                                                    pvc_uid=pvc_uid)
    except Exception as exception:
        logger.error(f'Creating Kubernetes event for {pvc_namespace}/{pvc_name} failed with a {exception.__class__.__name__} exception: {str(exception)}')
        pass


@ceph.signal_backup_post_success.connect
def ceph_backup_post_success(sender: str, volume: str, pool: str, image: str, version_labels: Dict[str, str],
                             context: Dict[str, Any], version: Optional[Dict]):
    assert isinstance(context, dict)
    assert version is not None

    pvc_namespace = context['pvc'].metadata.namespace
    pvc_name = context['pvc'].metadata.name
    pvc_uid = context['pvc'].metadata.uid
    start_time = context['backup-start-time']

    completion_time = time.time()
    prometheus.backup_completion_time.labels(volume=volume).set(completion_time)
    prometheus.backup_runtime_seconds.labels(volume=volume).set(completion_time - start_time)
    prometheus.backup_status_succeeded.labels(volume=volume).set(1)

    try:
        benji.k8s_tools.kubernetes.create_pvc_event(
            type='Normal',
            reason='SuccessfulBackup',
            message=f'Backup to {version["uid"]} completed successfully (took {completion_time - start_time:.0f} seconds).',
            pvc_namespace=pvc_namespace,
            pvc_name=pvc_name,
            pvc_uid=pvc_uid)
    except Exception as exception:
        logger.error(f'Creating Kubernetes event for {pvc_namespace}/{pvc_name} failed with a {exception.__class__.__name__} exception: {str(exception)}')
        pass


@ceph.signal_backup_post_error.connect
def ceph_backup_post_error(sender: str, volume: str, pool: str, image: str, version_labels: Dict[str, str],
                           context: Dict[str, Any], version: Optional[Dict], exception: Exception):
    assert isinstance(context, dict)
    pvc_namespace = context['pvc'].metadata.namespace
    pvc_name = context['pvc'].metadata.name
    pvc_uid = context['pvc'].metadata.uid

    start_time = context['backup-start-time']
    completion_time = time.time()

    prometheus.backup_completion_time.labels(volume=volume).set(completion_time)
    prometheus.backup_runtime_seconds.labels(volume=volume).set(completion_time - start_time)
    prometheus.backup_status_failed.labels(volume=volume).set(1)

    benji.k8s_tools.kubernetes.create_pvc_event(type='Warning',
                                                reason='FailedBackup',
                                                message=f'Backup failed: {exception.__class__.__name__} {str(exception)}',
                                                pvc_namespace=pvc_namespace,
                                                pvc_name=pvc_name,
                                                pvc_uid=pvc_uid)

    raise exception


def main():
    # This arguments parser tries to mimic kubectl
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter, allow_abbrev=False)

    parser.add_argument('-n',
                        '--namespace',
                        metavar='namespace',
                        dest='namespace',
                        default=None,
                        help='Filter on namespace')
    parser.add_argument('-l',
                        '--selector',
                        metavar='label-selector',
                        dest='labels',
                        action='append',
                        default=[],
                        help='Filter PVCs on label selector')
    parser.add_argument('--field-selector',
                        metavar='field-selector',
                        dest='fields',
                        action='append',
                        default=[],
                        help='Filter PVCs on field selector')

    args = parser.parse_args()

    benji.k8s_tools.kubernetes.load_config()
    core_v1_api = kubernetes.client.CoreV1Api()

    labels = ','.join(args.labels)
    fields = ','.join(args.fields)

    if args.namespace is not None:
        logger.info(f'Backing up all PVCs in namespace {args.namespace}.')
    else:
        logger.info(f'Backing up all PVCs in all namespaces.')
    if labels != '':
        logger.info(f'Matching label(s) {labels}.')
    if fields != '':
        logger.info(f'Matching field(s) {fields}.')

    if args.namespace is not None:
        pvcs = core_v1_api.list_namespaced_persistent_volume_claim(args.namespace,
                                                                   watch=False,
                                                                   label_selector=labels,
                                                                   field_selector=fields).items
    else:
        pvcs = core_v1_api.list_persistent_volume_claim_for_all_namespaces(watch=False,
                                                                           label_selector=labels,
                                                                           field_selector=fields).items
    if len(pvcs) == 0:
        logger.info('Not matching PVCs found.')
        sys.exit(0)

    for pvc in pvcs:
        if not hasattr(pvc.spec, 'volume_name') or pvc.spec.volume_name in (None, ''):
            continue

        pv = core_v1_api.read_persistent_volume(pvc.spec.volume_name)
        pool, image = benji.k8s_tools.kubernetes.determine_rbd_image_from_pv(pv)
        if pool is None or image is None:
            continue

        volume = f'{pvc.metadata.namespace}/{pvc.metadata.name}'
        # Limit the version_uid to 253 characters so that it is a compatible Kubernetes resource name.
        version_uid = '{}-{}'.format(f'{pvc.metadata.namespace}-{pvc.metadata.name}'[:246], _random_string(6))

        version_labels = {
            'benji-backup.me/instance': settings.benji_instance,
            'benji-backup.me/ceph-pool': pool,
            'benji-backup.me/ceph-rbd-image': image,
            'benji-backup.me/k8s-pvc-namespace': pvc.metadata.namespace,
            'benji-backup.me/k8s-pvc': pvc.metadata.name,
            'benji-backup.me/k8s-pv': pv.metadata.name
        }

        context = {'pvc': pvc}
        ceph.backup(volume=volume,
                    pool=pool,
                    image=image,
                    version_uid=version_uid,
                    version_labels=version_labels,
                    context=context)

    prometheus.push(prometheus.backup_registry)
    sys.exit(0)



"""bytes_dedup

Revision ID: 3d014d45493f
Revises: da25cc147a07
Create Date: 2019-10-28 19:52:25.464697

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '3d014d45493f'
down_revision = 'da25cc147a07'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('versions', schema=None) as batch_op:
        batch_op.alter_column('bytes_dedup',
                              new_column_name='bytes_deduplicated',
                              existing_type=sa.BigInteger(),
                              existing_nullable=True)


def downgrade():
    pass



import importlib
from typing import NamedTuple, Any, Dict
from urllib import parse

from benji.config import Config, ConfigList
from benji.exception import InternalError, ConfigurationError, UsageError
from benji.io.base import IOBase
from benji.repr import ReprMixIn


class _IOFactoryModule(NamedTuple):
    module: Any
    arguments: Dict[str, Any]


class IOFactory(ReprMixIn):

    _modules: Dict[str, _IOFactoryModule] = {}

    def __init__(self) -> None:
        raise InternalError('IOFactory constructor called.')

    @classmethod
    def _import_modules(cls, config: Config, modules: ConfigList) -> None:
        for index, module_dict in enumerate(modules):
            module = Config.get_from_dict(module_dict,
                                          'module',
                                          types=str,
                                          full_name_override=modules.full_name,
                                          index=index)
            name = Config.get_from_dict(module_dict,
                                        'name',
                                        types=str,
                                        full_name_override=modules.full_name,
                                        index=index)
            configuration = Config.get_from_dict(module_dict,
                                                 'configuration',
                                                 None,
                                                 types=dict,
                                                 full_name_override=modules.full_name,
                                                 index=index)

            if name in cls._modules:
                raise ConfigurationError('Duplicate name "{}" in list {}.'.format(name, modules.full_name))

            module = importlib.import_module('{}.{}'.format(__package__, module))
            try:
                configuration = config.validate(module=module.__name__, config=configuration)
            except ConfigurationError as exception:
                raise ConfigurationError('Configuration for IO {} is invalid.'.format(name)) from exception
            cls._modules[name] = _IOFactoryModule(module=module,
                                                  arguments={
                                                      'config': config,
                                                      'name': name,
                                                      'module_configuration': configuration
                                                  })

    @classmethod
    def initialize(cls, config: Config) -> None:
        ios: ConfigList = config.get('ios', None, types=list)
        cls._modules = {}
        cls._import_modules(config, ios)

    @classmethod
    def get(cls, url: str, block_size: int) -> IOBase:
        parsed_url = parse.urlparse(url)

        name = parsed_url.scheme
        if not name:
            raise UsageError('The supplied URL {} is invalid. You must provide a scheme.'.format(url))
        if name not in cls._modules:
            raise ConfigurationError('IO scheme {} is undefined.'.format(name))

        module = cls._modules[name].module
        module_arguments = cls._modules[name].arguments.copy()
        module_arguments['url'] = url
        module_arguments['block_size'] = block_size
        return module.IO(**module_arguments)



import os

running_pod_name = os.getenv('POD_NAME', 'unknown-pod-name')

benji_instance = os.getenv('BENJI_INSTANCE', 'benji-k8s')
benji_log_level = os.getenv('BENJI_LOG_LEVEL', 'INFO')
prom_push_gateway = os.getenv('PROM_PUSH_GATEWAY', None)



#!/usr/bin/env python3
import argparse
import logging
import sys
import time

import kubernetes
from kubernetes.client.rest import ApiException

import benji.helpers.settings as settings
import benji.helpers.utils as utils
import benji.k8s_tools.kubernetes

utils.setup_logging()
logger = logging.getLogger()


def main():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter, allow_abbrev=False)

    parser.add_argument('-f',
                        '--force',
                        dest='force',
                        action='store_true',
                        default=False,
                        help='Overwrite content of existing persistent volumes')
    parser.add_argument('--pvc-storage-class',
                        metavar='pvc_storage_class',
                        dest='pvc_storage_class',
                        default='rbd',
                        help='PVC storage class (only takes effect if the PVC does not exist yet)')
    parser.add_argument('--restore-url-template',
                        metavar='restore_url_template',
                        dest='restore_url_template',
                        help='Template to use for constructing URL for benji restore call',
                        default='rbd:{pool}/{image}')
    parser.add_argument(metavar='version_uid', dest='version_uid', help='Version uid')
    parser.add_argument(metavar='pvc_namespace', dest='pvc_namespace', help='PVC namespace')
    parser.add_argument(metavar='pvc_name', dest='pvc_name', help='PVC name')

    args = parser.parse_args()

    benji.k8s_tools.kubernetes.load_config()

    logger.info(f'Restoring version {args.version_uid} to PVC {args.pvc_namespace}/{args.pvc_name}.')

    benji_ls = utils.subprocess_run(
        ['benji', '--machine-output', '--log-level', settings.benji_log_level, 'ls', f'uid == "{args.version_uid}"'],
        decode_json=True)
    assert isinstance(benji_ls, dict)
    assert 'versions' in benji_ls
    assert isinstance(benji_ls['versions'], list)

    if len(benji_ls['versions']) == 0:
        raise RuntimeError(f'Size of {args.version_uid} could not be determined.')

    assert isinstance(benji_ls['versions'][0], dict)
    assert isinstance(benji_ls['versions'][0]['size'], int)
    version_size = benji_ls['versions'][0]['size']

    # This assumes that the Kubernetes client has already been initialized
    core_v1_api = kubernetes.client.CoreV1Api()
    pvc = None
    try:
        pvc = core_v1_api.read_namespaced_persistent_volume_claim(args.pvc_name, args.pvc_namespace)
    except ApiException as exception:
        if exception.status != 404:
            raise RuntimeError(f'Unexpected Kubernetes API exception: {str(exception)}')

    if pvc is None:
        pvc = benji.k8s_tools.kubernetes.create_pvc(args.pvc_name, args.pvc_namespace, version_size,
                                                    args.pvc_storage_class)
    else:
        if not args.force:
            raise RuntimeError('PVC already exists. Will not overwrite it unless forced.')

        # I don't really understand why capacity is a regular dict and not an object. Oh, well.
        pvc_size = int(benji.k8s_tools.kubernetes.parse_quantity(pvc.status.capacity['storage']))
        if pvc_size < version_size:
            raise RuntimeError(f'Existing PVC is too small to hold version {args.version_uid} ({pvc_size} < {version_size}).')
        elif pvc_size > version_size:
            logger.warning(f'Existing PVC is {pvc_size - version_size} bytes bigger than version {args.version_uid}.')

    while True:
        pvc = core_v1_api.read_namespaced_persistent_volume_claim(args.pvc_name, args.pvc_namespace)
        if pvc.status.phase == 'Bound':
            break
        logger.info('Waiting for persistent volume creation.')
        time.sleep(1)

    pv = core_v1_api.read_persistent_volume(pvc.spec.volume_name)
    pool, image = benji.k8s_tools.kubernetes.determine_rbd_image_from_pv(pv)
    if pool is None or image is None:
        raise RuntimeError(f'Unable to determine PersistentVolume pool or image for {pv.metadata.name}')

    utils.subprocess_run([
        'benji',
        '--log-level',
        settings.benji_log_level,
        'restore',
        '--sparse',
        '--force',
        args.version_uid,
        args.restore_url_template.format(pool=pool, image=image),
    ])
    sys.exit(0)



"""update_20190118-1

Revision ID: 2e028f08346b
Revises: 
Create Date: 2019-01-18 16:54:13.525607

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '2e028f08346b'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('blocks', schema=None) as batch_op:
        batch_op.create_index(batch_op.f('ix_blocks_uid_left'), ['uid_left', 'uid_right'], unique=False)
        batch_op.drop_index('ix_blocks_uid_left_uid_right')

    with op.batch_alter_table('deleted_blocks', schema=None) as batch_op:
        batch_op.create_index(batch_op.f('ix_deleted_blocks_uid_left'), ['uid_left', 'uid_right'], unique=False)
        batch_op.drop_index('ix_blocks_uid_left_uid_right_2')

    with op.batch_alter_table('labels', schema=None) as batch_op:
        batch_op.alter_column('name', existing_type=sa.VARCHAR(), type_=sa.String(length=255))
        batch_op.alter_column('value', existing_type=sa.VARCHAR(), type_=sa.String(length=255), existing_nullable=False)
        batch_op.create_unique_constraint(batch_op.f('uq_labels_version_uid'), ['version_uid', 'name'])

    with op.batch_alter_table('locks', schema=None) as batch_op:
        batch_op.alter_column('host', existing_type=sa.VARCHAR(), type_=sa.String(length=255))
        batch_op.alter_column('lock_name', existing_type=sa.VARCHAR(), type_=sa.String(length=255))
        batch_op.alter_column('process_id', existing_type=sa.VARCHAR(), type_=sa.String(length=255))
        batch_op.alter_column(
            'reason', existing_type=sa.VARCHAR(), type_=sa.String(length=255), existing_nullable=False)

    with op.batch_alter_table('version_statistics', schema=None) as batch_op:
        batch_op.alter_column('name', existing_type=sa.VARCHAR(), type_=sa.String(length=255), existing_nullable=False)
        batch_op.alter_column(
            'snapshot_name', existing_type=sa.VARCHAR(), type_=sa.String(length=255), existing_nullable=False)

    with op.batch_alter_table('versions', schema=None) as batch_op:
        batch_op.alter_column('name', existing_type=sa.VARCHAR(), type_=sa.String(length=255), existing_nullable=False)
        batch_op.alter_column(
            'snapshot_name',
            existing_type=sa.VARCHAR(),
            server_default=None,
            type_=sa.String(length=255),
            existing_nullable=False)


def downgrade():
    pass



import datetime
import json
import logging
import re
import time
import uuid
from subprocess import TimeoutExpired, CalledProcessError
from typing import List, Union, Tuple, Optional

import kubernetes
from kubernetes.stream import stream
from kubernetes.stream.ws_client import ERROR_CHANNEL, STDOUT_CHANNEL, STDERR_CHANNEL

from benji.helpers.settings import running_pod_name, benji_instance

SERVICE_NAMESPACE_FILENAME = '/var/run/secrets/kubernetes.io/serviceaccount/namespace'

logger = logging.getLogger()


def load_config() -> None:
    try:
        kubernetes.config.load_incluster_config()
        logger.debug('Configured in cluster with service account.')
    except Exception:
        try:
            kubernetes.config.load_kube_config()
            logger.debug('Configured via kubeconfig file.')
        except Exception:
            raise RuntimeError('No Kubernetes configuration found.')


def service_account_namespace() -> str:
    with open(SERVICE_NAMESPACE_FILENAME, 'r') as f:
        namespace = f.read()
        if namespace == '':
            raise RuntimeError(f'{SERVICE_NAMESPACE_FILENAME} is empty.')
    return namespace


# This was implemented with version 10.0.0 of the Python kubernetes client in mind. But there are several open issues
# and PRs regarding encoding and timeout with might affect us in the future:
#
#  https://github.com/kubernetes-client/python-base/issues/106
#  https://github.com/kubernetes-client/python-base/pull/143
#  https://github.com/kubernetes-client/python-base/pull/78
#
# kubectl uses a POST request to establish the pod connection. We mimic this here by using
# connect_post_namespaced_pod_exec. The examples from the kubernetes client use connect_get_namespaced_pod_exec instead.
# There shouldn't be any differences in functionality but the settings in the RBAC role are different (create vs. get)
# which is why we follow the kubectl implementation here.
def pod_exec(args: List[str],
             *,
             name: str,
             namespace: str,
             container: str = None,
             timeout: float = float("inf")) -> Tuple[str, str]:
    core_v1_api = kubernetes.client.CoreV1Api()
    logger.debug('Running command in pod {}/{}: {}.'.format(namespace, name, ' '.join(args)))
    ws_client = stream(core_v1_api.connect_post_namespaced_pod_exec,
                       name,
                       namespace,
                       command=args,
                       container=container,
                       stderr=True,
                       stdin=False,
                       stdout=True,
                       tty=False,
                       _preload_content=False)

    start = time.time()
    while ws_client.is_open() and time.time() - start < timeout:
        ws_client.update(timeout=(timeout - time.time() + start))

    stdout_channel = ws_client.read_channel(STDOUT_CHANNEL, timeout=0)
    stderr_channel = ws_client.read_channel(STDERR_CHANNEL, timeout=0)
    error_channel = ws_client.read_channel(ERROR_CHANNEL, timeout=0)
    ws_client.close()
    if error_channel == '':
        raise TimeoutExpired(cmd=args, timeout=timeout, output=stdout_channel, stderr=stderr_channel)
    else:
        error_channel_object = json.loads(error_channel)

        # Failure example:
        # {
        #   "metadata": {},
        #   "status": "Failure",
        #   "message": "command terminated with non-zero exit code: Error executing in Docker Container: 126",
        #   "reason": "NonZeroExitCode",
        #   "details": {
        #     "causes": [
        #       {
        #         "reason": "ExitCode",
        #         "message": "126"
        #       }
        #     ]
        #   }
        # }
        #
        # Non-zero exit codes from the command ran are also returned this way.
        #
        # Success example:
        # {"metadata":{},"status":"Success"}
        #
        # See: https://github.com/kubernetes/kubernetes/blob/87b744715ec6952c45d04253dc7b63fc3cfe1ddc/staging/src/k8s.io/client-go/tools/remotecommand/v4.go#L82
        #      https://github.com/kubernetes-client/python-base/blob/master/stream/ws_client.py

        assert isinstance(error_channel_object, dict)
        assert 'status' in error_channel_object
        if error_channel_object['status'] == 'Success':
            pass
        elif error_channel_object['status'] == 'Failure' and 'reason' in error_channel_object and error_channel_object['reason'] == 'NonZeroExitCode':
            assert 'details' in error_channel_object
            assert 'causes' in error_channel_object['details']
            assert isinstance(error_channel_object['details']['causes'], list)
            for cause in error_channel_object['details']['causes']:
                assert 'reason' in cause
                if cause['reason'] != 'ExitCode':
                    continue
                assert 'message' in cause
                raise CalledProcessError(returncode=int(cause["message"]),
                                         cmd=args,
                                         output=stdout_channel,
                                         stderr=stderr_channel)
        else:
            raise RuntimeError(f'Unknown stream status: {error_channel_object["status"]}/{error_channel_object.get("reason", "mot-set")}.')

    return stdout_channel, stderr_channel


def create_pvc_event(*, type: str, reason: str, message: str, pvc_namespace: str, pvc_name: str,
                     pvc_uid: str) -> kubernetes.client.models.v1_event.V1Event:
    event_name = '{}-{}'.format(benji_instance, str(uuid.uuid4()))
    # Kubernetes requires a time including microseconds
    event_time = datetime.datetime.utcnow().isoformat(timespec='microseconds') + 'Z'

    # Setting uid is required so that kubectl describe finds the event.
    # And setting firstTimestamp is required so that kubectl shows a proper age for it.
    # See: https://github.com/kubernetes/kubernetes/blob/
    event = {
        'apiVersion': 'v1',
        'kind': 'Event',
        'metadata': {
            'name': event_name,
            'namespace': pvc_namespace,
            'labels': {
                'reporting-component': 'benji',
                'reporting-instance': running_pod_name,
                'benji-backup.me/instance': benji_instance
            }
        },
        'involvedObject': {
            'apiVersion': 'v1',
            'kind': 'PersistentVolumeClaim',
            'name': pvc_name,
            'namespace': pvc_namespace,
            'uid': pvc_uid
        },
        'eventTime': event_time,
        'firstTimestamp': event_time,
        'lastTimestamp': event_time,
        'type': type,
        'reason': reason,
        # Message can be at most 1024 characters long
        'message': message[:1024],
        'action': 'None',
        'reportingComponent': 'benji',
        'reportingInstance': running_pod_name,
        'source': {
            'component': 'benji'
        }
    }

    core_v1_api = kubernetes.client.CoreV1Api()
    return core_v1_api.create_namespaced_event(namespace=pvc_namespace, body=event)


def create_pvc(pvc_name: str, pvc_namespace: int, pvc_size: str,
               pvc_storage_class: str) -> kubernetes.client.models.v1_persistent_volume_claim.V1PersistentVolumeClaim:
    pvc = {
        'kind': 'PersistentVolumeClaim',
        'apiVersion': 'v1',
        'metadata': {
            'namespace': pvc_namespace,
            'name': pvc_name,
        },
        'spec': {
            'storageClassName': pvc_storage_class,
            'accessModes': ['ReadWriteOnce'],
            'resources': {
                'requests': {
                    'storage': pvc_size
                }
            }
        }
    }

    core_v1_api = kubernetes.client.CoreV1Api()
    return core_v1_api.create_namespaced_persistent_volume_claim(namespace=pvc_namespace, body=pvc)


def determine_rbd_image_from_pv(
        pv: kubernetes.client.models.v1_persistent_volume.V1PersistentVolume) -> Tuple[Optional[str], Optional[str]]:
    pool, image = None, None

    if hasattr(pv.spec, 'rbd') and hasattr(pv.spec.rbd, 'pool') and hasattr(pv.spec.rbd, 'image'):
        # Native Kubernetes RBD PV
        pool, image = pv.spec.rbd.pool, pv.spec.rbd.image
    elif hasattr(pv.spec, 'flex_volume') and hasattr(pv.spec.flex_volume, 'options') and hasattr(
            pv.spec.flex_volume, 'driver'):
        # Rook Ceph PV
        options = pv.spec.flex_volume.options
        driver = pv.spec.flex_volume.driver
        if driver.startswith('ceph.rook.io/') and options.get('pool') and options.get('image'):
            pool, image = options['pool'], options['image']
    elif (hasattr(pv.spec, 'csi') and hasattr(pv.spec.csi, 'driver') and
          pv.spec.csi.driver in ('rook-ceph.rbd.csi.ceph.com', 'rbd.csi.ceph.com') and
          hasattr(pv.spec.csi, 'volume_handle') and pv.spec.csi.volume_handle and
          hasattr(pv.spec.csi, 'volume_attributes') and pv.spec.csi.volume_attributes.get('pool')):
        attributes = pv.spec.csi.volume_attributes
        volume_handle_parts = pv.spec.csi.volume_handle.split('-')
        if len(volume_handle_parts) >= 9:
            image_prefix = attributes.get('volumeNamePrefix', 'csi-vol-')
            image_suffix = '-'.join(volume_handle_parts[len(volume_handle_parts) - 5:])
            pool, image = attributes['pool'], image_prefix + image_suffix

    return pool, image


# This is taken from https://github.com/kubernetes-client/python/pull/855 with minimal changes.
#
# Copyright 2019 The Kubernetes Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def parse_quantity(quantity: Union[str, int, float]) -> float:
    """
    Parse kubernetes canonical form quantity like 200Mi to an float number.
    Supported SI suffixes:
    base1024: Ki | Mi | Gi | Ti | Pi | Ei
    base1000: m | "" | k | M | G | T | P | E

    Input:
    quanity: string. kubernetes canonical form quantity

    Returns:
    float

    Raises:
    ValueError on invalid or unknown input
    """
    exponents = {"m": -1, "K": 1, "k": 1, "M": 2, "G": 3, "T": 4, "P": 5, "E": 6}
    pattern = r"^(\d+)([^\d]{1,2})?$"

    if isinstance(quantity, (int, float)):
        return float(quantity)

    quantity = str(quantity)

    res = re.match(pattern, quantity)
    if not res:
        raise ValueError("{} did not match pattern {}".format(quantity, pattern))
    number, suffix = res.groups()
    number_float = float(number)

    if suffix is None:
        return number_float

    suffix = res.groups()[1]

    if suffix.endswith("i"):
        base = 1024
    elif len(suffix) == 1:
        base = 1000
    else:
        raise ValueError("{} has unknown suffix".format(quantity))

    # handle SI inconsistency
    if suffix == "ki":
        raise ValueError("{} has unknown suffix".format(quantity))

    if suffix[0] not in exponents:
        raise ValueError("{} has unknown suffix".format(quantity))

    exponent = exponents[suffix[0]]
    return number_float * (base**exponent)


# End of included content
